"""
Draws the graph into an image.
Input format: PNG / CSV
Output format: PNG

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
from tqdm import tqdm
import os
import cv2
import numpy as np
import matplotlib.pyplot as plt
import json
from typing import Dict, Tuple, List
from concurrent.futures import ThreadPoolExecutor
import dgl
from dgl import DGLHeteroGraph
import networkx as nx
from ..utils.preprocessing import get_names
from ..preprocessing import pngcsv2graph_main
from ..classification.read_graph import GraphDataset
from ..utils.read_nodes import read_node_matrix


def get_colors(type_info: Dict[str, Tuple[str, List[int]]]) -> List[str]:
    """
    Retrieves a list of hexadecimal colors from a dictionary of RGB tuples.

    :param type_info: A dictionary containing type information as keys and RGB tuples as values.
    :type type_info: Dict[str, Tuple[str, List[int]]]
    
    :return: A list of hexadecimal colors converted from the RGB tuples.
    :rtype: List[str]
    """
    def to_hex(rgb_tuple: Tuple[int, int, int]) -> str:
        """
        Converts an RGB tuple to a hexadecimal color representation.

        :param rgb_tuple: A tuple containing three integer values representing RGB channels.
        :type rgb_tuple: Tuple[int, int, int]

        :return: A string representing the hexadecimal color code.
        :rtype: str
        """
        hex_values = [format(value, '02x') for value in rgb_tuple]
        return '#' + ''.join(hex_values)
    return [to_hex(rgb_tuple) for name, rgb_tuple in type_info.values()]


def draw_graph(
        orig: np.ndarray,
        graph: DGLHeteroGraph,
        x: np.ndarray,
        y: np.ndarray,
        labels: np.ndarray,
        type_info: Dict[str, Tuple[str, List[int]]],
        save_path: str
        ) -> np.ndarray:
    """
    Draws a graph into an image using Matplotlib and NetworkX.

    :param orig: The original data used to construct the graph.
    :type orig: np.ndarray

    :param graph: The graph object to be visualized.
    :type graph: DGLHeteroGraph

    :param x: Node features used for plotting.
    :type x: np.ndarray

    :param y: Node coordinates used for plotting.
    :type y: np.ndarray

    :param labels: Node labels used for visualization.
    :type labels: np.ndarray

    :param type_info: A dictionary containing type information as keys and RGB tuples as values.
    :type type_info: Dict[str, Tuple[str, List[int]]]

    :param save_path: The file path to save the resulting graph visualization.
    :type save_path: str

    :return: The image of the graph.
    :rtype: np.ndarray
    """
    gx = dgl.to_networkx(graph)
    pos = {k: (y[k], x[k]) for k in range(len(x))}
    colors = get_colors(type_info)
    cols = [colors[label + 1] for label in labels]
    fig = plt.figure(frameon=True)
    fig.set_size_inches(4,4)
    fig.tight_layout()
    ax = plt.Axes(fig, [0., 0., 1., 1.])
    ax.set_axis_off()
    fig.add_axes(ax)
    nx.draw_networkx(gx, pos=pos, arrows=False, node_color=cols, with_labels=False, node_size=10, ax=ax)
    ax.imshow(orig, aspect='auto')
    extent = ax.get_window_extent().transformed(fig.dpi_scale_trans.inverted())
    fig.savefig(save_path, dpi=256, bbox_inches=extent)
    return orig


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--orig-dir', type=str, help='Path to base images folder. Must be in .png format.')
    parser.add_argument('--png-dir', type=str, help='Path to folder with png of the labels.')
    parser.add_argument('--csv-dir', type=str, help='Path to folder with csv of the labels.')
    parser.add_argument('--output-dir', type=str, help='Path to folder where to save results.')
    parser.add_argument('--type-info', type=str, help='Path to type_info.json.')
    parser.add_argument('--num-workers', type=int, default=0)
    parser.add_argument('--max-degree', type=int, default=10, help='Maximum degree allowed for each node.')
    parser.add_argument('--max-distance', type=int, default=200, help='Maximum allowed distance between nodes, in pixels.')
    return parser


def main_subthread(
        args: Namespace,
        name: str,
        graph_dataset: GraphDataset,
        type_info: Dict[str, Tuple[str, List[int]]],
        k: int,
        pbar: tqdm,
        ):
    orig = cv2.imread(os.path.join(args.orig_dir, name + '.png'), cv2.IMREAD_COLOR)[:, :, ::-1]
    X, y, xx, yy = read_node_matrix(
        os.path.join(args.output_dir, 'graphs', name + '.nodes.csv'), return_coordinates=True, return_class=True,
        remove_prior=False, remove_morph=False
        )
    graph = graph_dataset[k]
    draw_graph(orig, graph, xx, yy, y, type_info, os.path.join(args.output_dir, name + '.graph-overlay.png'))
    nx_graph = dgl.to_networkx(graph)
    nodes = [
        (i, {'target': str(y[i])}) for i in nx_graph.nodes()
    ]
    nx_graph.add_nodes_from(nodes)
    nx.write_gml(nx_graph, os.path.join(args.output_dir, 'graphs', name + '.gml'))
    pbar.update(1)


def main_with_args(args: Namespace) -> None:
    os.makedirs(args.output_dir, exist_ok=True)
    newargs = Namespace(
        png_dir=args.png_dir,
        csv_dir=args.csv_dir,
        orig_dir=args.orig_dir,
        output_path=os.path.join(args.output_dir, 'graphs'),
        num_workers=args.num_workers
    )
    pngcsv2graph_main(newargs)
    graph_dataset = GraphDataset(
        node_dir = os.path.join(args.output_dir, 'graphs'),
        max_degree = args.max_degree,
        max_dist = args.max_distance
    )
    names = sorted(get_names(args.orig_dir, '.png'))
    with open(args.type_info, 'r') as f:
        type_info = json.load(f)
    pbar = tqdm(total=len(names))
    plt.switch_backend('Agg')
    if args.num_workers > 0:
        with ThreadPoolExecutor(max_workers=args.num_workers) as executor:
            for k, name in enumerate(names):
                executor.submit(main_subthread, args, name, graph_dataset, type_info, k, pbar)
    else:
        for k, name in enumerate(names):
            main_subthread(args, name, graph_dataset, type_info, k, pbar)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Converts our png <-> csv format into QuPath geojson format.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import List, Dict, Any, Optional
import argparse
import pandas as pd
import cv2
import numpy as np
from tqdm import tqdm
from ..utils.preprocessing import (
    get_names, create_dir, parse_path,
    format_contour, create_geojson, read_labels,
    save_geojson
)


def create_mask(png: np.ndarray, csv: pd.DataFrame, label: int) -> np.ndarray:
    """
    Returns an image with only the pixels of the class given in the label.

    :param png: Segmentation mask with indices as pixel values.
    :type png: np.ndarray
    :param csv: DataFrame containing the class labels for each cell in the image.
    :type csv: pd.DataFrame
    :param label: Class label to extract from the image.
    :type label: int
    :return: Binary mask where pixels of the specified class have value 1 and all others have value 0.
    :rtype: np.ndarray
    """
    mask = png.copy()
    for i, (idx, cell_label) in csv.iterrows():
        if cell_label != label:
            mask[mask == idx] = 0
    return np.array(mask, dtype=np.uint8)


def pngcsv2features(png: np.ndarray, csv: pd.DataFrame, label: int, num_classes: Optional[int] = 2) -> List[Dict[str, Any]]:
    """
    Computes geojson features of contours of a given class.

    :param png: Segmentation mask with indices as pixel values.
    :type png: np.ndarray
    :param csv: DataFrame with the classes of the cells in the image.
    :type csv: pd.DataFrame
    :param label: The label of the class of cells to extract features for.
    :type label: int
    :param num_classes: The number of classes in the segmentation mask, defaults to 2.
    :type num_classes: Optional[int], optional
    :return: A list of dictionaries with the geojson features of the contours of the specified class of cells.
    :rtype: List[Dict[str, Any]]
    """
    mask = create_mask(png, csv, label)
    contours, _ = cv2.findContours(mask, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_SIMPLE)
    contours = filter(lambda x: len(x[0]) >= 4, [(format_contour(c), label) for c in contours])
    return create_geojson(contours, num_classes)


def pngcsv2geojson(png: np.ndarray, csv: pd.DataFrame, num_classes: Optional[int] = 2) -> List[Dict[str, Any]]:
    """
    Computes the GeoJSON representation of contours in the given PNG image based on the provided CSV labels.

    :param png: A NumPy array representing the PNG image.
    :type png: np.ndarray
    :param csv: A Pandas DataFrame with the CSV labels of each cell.
    :type csv: pd.DataFrame
    :param num_classes: The number of classes. Default is 2.
    :type num_classes: Optional[int]
    :return: A list of features representing contours.
    :rtype: List[Dict[str, Any]]
    """
    total_contours = []
    for _, (idx, cell_label) in csv.iterrows():
        mask = png.copy()
        mask[mask != idx] = 0
        mask[mask == idx] = 1
        mask = mask.astype(np.uint8)
        contours, _ = cv2.findContours(mask, mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_SIMPLE)
        contours = filter(lambda x: len(x[0]) >= 4, [(format_contour(c), cell_label) for c in contours])
        total_contours.extend(create_geojson(contours, num_classes))
        del mask
    return total_contours


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--png-dir', type=str, required=True,
                        help='Path to png files.')
    parser.add_argument('--csv-dir', type=str, required=True,
                        help='Path to csv files.')
    parser.add_argument('--gson-dir', type=str, required=True,
                        help='Path to save files.')
    parser.add_argument('--num-classes', type=int, default=2)
    return parser


def main_with_args(args):
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)
    output_path = parse_path(args.gson_dir)
    create_dir(output_path)

    names = get_names(png_dir, '.GT_cells.png')
    for name in tqdm(names):
        png, csv = read_labels(name, png_dir, csv_dir)
        if png is None or png.max() == 0:
            continue
        gson = pngcsv2geojson(png, csv, args.num_classes)
        save_geojson(gson, name, output_path)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



import math
import cv2
import matplotlib.pyplot as plt
import numpy as np

from .utils import cropping_center


#####
class PatchExtractor(object):
    """Extractor to generate patches with or without padding.
    Turn on debug mode to see how it is done.

    Args:
        x         : input image, should be of shape HWC
        win_size  : a tuple of (h, w)
        step_size : a tuple of (h, w)
        debug     : flag to see how it is done
    Return:
        a list of sub patches, each patch has dtype same as x

    Examples:
        >>> xtractor = PatchExtractor((450, 450), (120, 120))
        >>> img = np.full([1200, 1200, 3], 255, np.uint8)
        >>> patches = xtractor.extract(img, 'mirror')

    """

    def __init__(self, win_size, step_size, debug=False):

        self.patch_type = "mirror"
        self.win_size = win_size
        self.step_size = step_size
        self.debug = debug
        self.counter = 0

    def __get_patch(self, x, ptx):
        pty = (ptx[0] + self.win_size[0], ptx[1] + self.win_size[1])
        win = x[ptx[0] : pty[0], ptx[1] : pty[1]]
        assert (
            win.shape[0] == self.win_size[0] and win.shape[1] == self.win_size[1]
        ), "[BUG] Incorrect Patch Size {0}".format(win.shape)
        if self.debug:
            if self.patch_type == "mirror":
                cen = cropping_center(win, self.step_size)
                cen = cen[..., self.counter % 3]
                cen.fill(150)
            cv2.rectangle(x, ptx, pty, (255, 0, 0), 2)
            plt.imshow(x)
            plt.show(block=False)
            plt.pause(1)
            plt.close()
            self.counter += 1
        return win

    def __extract_valid(self, x):
        """Extracted patches without padding, only work in case win_size > step_size.
        
        Note: to deal with the remaining portions which are at the boundary a.k.a
        those which do not fit when slide left->right, top->bottom), we flip 
        the sliding direction then extract 1 patch starting from right / bottom edge. 
        There will be 1 additional patch extracted at the bottom-right corner.

        Args:
            x         : input image, should be of shape HWC
            win_size  : a tuple of (h, w)
            step_size : a tuple of (h, w)
        Return:
            a list of sub patches, each patch is same dtype as x

        """
        im_h = x.shape[0]
        im_w = x.shape[1]

        def extract_infos(length, win_size, step_size):
            flag = (length - win_size) % step_size != 0
            last_step = math.floor((length - win_size) / step_size)
            last_step = (last_step + 1) * step_size
            return flag, last_step

        h_flag, h_last = extract_infos(im_h, self.win_size[0], self.step_size[0])
        w_flag, w_last = extract_infos(im_w, self.win_size[1], self.step_size[1])

        sub_patches = []
        #### Deal with valid block
        for row in range(0, h_last, self.step_size[0]):
            for col in range(0, w_last, self.step_size[1]):
                win = self.__get_patch(x, (row, col))
                sub_patches.append(win)
        #### Deal with edge case
        if h_flag:
            row = im_h - self.win_size[0]
            for col in range(0, w_last, self.step_size[1]):
                win = self.__get_patch(x, (row, col))
                sub_patches.append(win)
        if w_flag:
            col = im_w - self.win_size[1]
            for row in range(0, h_last, self.step_size[0]):
                win = self.__get_patch(x, (row, col))
                sub_patches.append(win)
        if h_flag and w_flag:
            ptx = (im_h - self.win_size[0], im_w - self.win_size[1])
            win = self.__get_patch(x, ptx)
            sub_patches.append(win)
        return sub_patches

    def __extract_mirror(self, x):
        """Extracted patches with mirror padding the boundary such that the 
        central region of each patch is always within the orginal (non-padded)
        image while all patches' central region cover the whole orginal image.

        Args:
            x         : input image, should be of shape HWC
            win_size  : a tuple of (h, w)
            step_size : a tuple of (h, w)
        Return:
            a list of sub patches, each patch is same dtype as x
            
        """
        diff_h = self.win_size[0] - self.step_size[0]
        padt = diff_h // 2
        padb = diff_h - padt

        diff_w = self.win_size[1] - self.step_size[1]
        padl = diff_w // 2
        padr = diff_w - padl

        pad_type = "constant" if self.debug else "reflect"
        x = np.lib.pad(x, ((padt, padb), (padl, padr), (0, 0)), pad_type)
        sub_patches = self.__extract_valid(x)
        return sub_patches

    def extract(self, x, patch_type):
        patch_type = patch_type.lower()
        self.patch_type = patch_type
        if patch_type == "valid":
            return self.__extract_valid(x)
        elif patch_type == "mirror":
            return self.__extract_mirror(x)
        else:
            assert False, "Unknown Patch Type [%s]" % patch_type
        return


# ----------------------------------------------------------------------------

if __name__ == "__main__":
    # toy example for debug
    # 355x355, 480x480
    xtractor = PatchExtractor((450, 450), (120, 120), debug=True)
    a = np.full([1200, 1200, 3], 255, np.uint8)
    xtractor.extract(a, "mirror")
    xtractor.extract(a, "valid")



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import pandas as pd
from sklearn.metrics import confusion_matrix

from tumourkit.utils.preprocessing import get_names, parse_path
from tumourkit.segmentation.evaluate import get_pairs

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
CENTROIDS_DIR = parse_path(TEST_DIR) + 'centroids/'


"""
The format of the names should be [name].A.csv and [name].B.csv,
being A and B two files to compare the metric. And [name].result.csv
should be the expected result. 
Labels should be only 1 and 2.
CSV files must have headers, except for [name].result.csv which should
only be the confusion matrix.
"""
@pytest.mark.parametrize("name", get_names(CENTROIDS_DIR, '.result.csv'))
def test_metric_pairs(name):
    A_centroids = pd.read_csv(CENTROIDS_DIR + name + '.A.csv').to_numpy()
    B_centroids = pd.read_csv(CENTROIDS_DIR + name + '.B.csv').to_numpy()
    result = pd.read_csv(CENTROIDS_DIR + name + '.result.csv', header=None).to_numpy()
    true, pred = get_pairs(A_centroids, B_centroids)
    conf_matrix = confusion_matrix(true, pred, labels=[0,1])
    assert((result == conf_matrix).all())



"""
Extract HoVer-Net probabilities from json and
concatenates the result as new column (prob1) to .nodes.csv.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from tqdm import tqdm
from typing import Dict, Any, List, Tuple, Optional
import pandas as pd
import numpy as np
import argparse
from argparse import Namespace
import os
from ..utils.preprocessing import create_dir, parse_path, get_names, read_json, save_graph
from ..utils.nearest import generate_tree, find_nearest
import logging
from logging import Logger


def parse_centroids_probs(nuc: Dict[str, Any], logger: Optional[Logger] = None, num_classes: Optional[int] = 2) -> List[Tuple[int, int, int]]:
    """
    Parses the centroids and probabilities from the Hovernet JSON nuclei dictionary.

    This function takes the Hovernet JSON nuclei dictionary as input, which is obtained from the modified run_infer.py script.
    It extracts the centroids and probabilities from the dictionary and returns them as a list of tuples.
    Each tuple represents a centroid and consists of (X, Y, prob1) for binary classification or (X, Y, prob1, prob2, ..., probN) for multiclass classification.

    :param nuc: The Hovernet JSON nuclei dictionary.
    :type nuc: Dict[str, Any]
    :param logger: Optional logger object to log warnings.
    :type logger: Optional[Logger]
    :param num_classes: Optional number of classes. Defaults to 2 for binary classification.
    :type num_classes: Optional[int]
    :return: The list of centroids with probabilities.
    :rtype: List[Tuple[int, int, int]]
    """
    centroids_ = []
    for inst in nuc:
        inst_info = nuc[inst]
        inst_centroid = inst_info['centroid']
        inst_prob1 = inst_info['prob1']
        if num_classes > 2:
            inst_probs = []
            for k in range(1, num_classes + 1):
                inst_probs.append(inst_info['prob' + str(k)])
        inst_type = inst_info['type']
        if inst_type == 0:
            if logger is None:
                logging.warning('Found cell with class 0, removing it.')
            else:
                logger.warning('Found cell with class 0, removing it.')
        else:
            if num_classes == 2:
                centroids_.append((inst_centroid[1], inst_centroid[0], inst_prob1))
            else:
                centroids_.append((inst_centroid[1], inst_centroid[0], *inst_probs))
    return centroids_


def add_probability(graph: pd.DataFrame, hov_json: Dict[str, Any], logger: Optional[Logger] = None, num_classes: Optional[int] = 2) -> pd.DataFrame:
    """
    Adds probability information from the Hovernet JSON nuclei dictionary to the graph DataFrame.

    This function extracts the type probabilities from the Hovernet JSON nuclei dictionary (`hov_json`) and adds them as columns to the `graph` DataFrame.
    The join between the `graph` DataFrame and the `hov_json` dictionary is based on the 'id' field.

    :param graph: The graph DataFrame containing the nodes.
    :type graph: pd.DataFrame
    :param hov_json: The Hovernet JSON nuclei dictionary.
    :type hov_json: Dict[str, Any]
    :param logger: Optional logger object to log warnings.
    :type logger: Optional[Logger]
    :param num_classes: Optional number of classes. Defaults to 2 for binary classification.
    :type num_classes: Optional[int]
    :return: The updated graph DataFrame with probability information.
    :rtype: pd.DataFrame
    """
    centroids = parse_centroids_probs(hov_json, logger, num_classes)
    centroids = np.array(centroids)
    assert len(centroids) > 0, 'Hov json must contain at least one cell.'
    graph = graph.copy()
    if 'prob1' not in graph.columns:
        n_cols = len(graph.columns)
        graph.insert(n_cols, 'prob1', [-1] * len(graph))
    else:
        graph['prob1'] = -1
    if num_classes > 2:
        for k in range(2, num_classes + 1):
            if not 'prob' + str(k) in graph.columns:
                n_cols = len(graph.columns)
                graph.insert(n_cols, 'prob' + str(k), [-1] * len(graph))
            else:
                graph['prob' + str(k)] = -1
    gt_tree = generate_tree(centroids[:, :2])
    pred_centroids = graph[['X', 'Y']].to_numpy(dtype=int)
    pred_tree = generate_tree(pred_centroids)
    for point_id, point in enumerate(centroids):
        closest_id = find_nearest(point[:2], pred_tree)
        closest = graph.loc[closest_id, ['X', 'Y', 'prob1']]
        if point_id == find_nearest(closest[:2], gt_tree):
            graph.loc[closest_id, 'prob1'] = point[2]  # 1-1 matchings
            if num_classes > 2:
                for k in range(2, num_classes + 1):
                    graph.loc[closest_id, 'prob' + str(k)] = point[k + 1]  # 1-1 matchings
    graph.drop(graph[graph['prob1'] == -1].index, inplace=True)
    return graph


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--json-dir', type=str, required=True,
        help='Path to folder containing HoVer-Net json outputs.'
    )
    parser.add_argument(
        '--graph-dir', type=str, required=True,
        help='Path to directory to .nodes.csv containing graph information.'
    )
    parser.add_argument(
        '--output-dir', type=str, required=True,
        help='Path where to save new .nodes.csv. If same as --graph-dir, overwrites its content.'
    )
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main_with_args(args: Namespace, logger: Optional[Logger] = None) -> None:
    json_dir = parse_path(args.json_dir)
    graph_dir = parse_path(args.graph_dir)
    output_dir = parse_path(args.output_dir)
    create_dir(output_dir)
    names = get_names(graph_dir, '.nodes.csv')
    for name in tqdm(names):
        try:
            graph = pd.read_csv(os.path.join(graph_dir, name + '.nodes.csv'))
            hov_json = read_json(os.path.join(json_dir, name + '.json'))
            graph = add_probability(graph, hov_json, logger, args.num_classes)
            save_graph(graph, os.path.join(output_dir, name + '.nodes.csv'))
        except FileNotFoundError:
            continue
    return


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Pipeline for inference.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
import logging
from logging import Logger
from .segmentation import hov_infer
import os
from .preprocessing import geojson2pngcsv_main, png2graph_main, hovernet2geojson_main, graph2centroids_main, centroidspng2csv_main, pngcsv2geojson_main
from .postprocessing import join_hovprob_graph_main
from .utils.preprocessing import create_dir
from .classification import infer_gnn
import pandas as pd


def set_best_configuration(args: Namespace, logger: Logger) -> None:
    """
    Sets the best configuration from training based on the F1 score.

    :param args: The arguments for setting the best configuration.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Configuration not provided, using best configuration from training based on F1 score.')
    if args.best_arch == 'GCN':
        save_file = os.path.join(args.root_dir, 'gnn_logs', 'gcn_results.csv')
    elif args.best_arch == 'ATT':
        save_file = os.path.join(args.root_dir, 'gnn_logs', 'gat_results.csv')
    else:
        assert False, 'Architecture not supported'
    gnn_results = pd.read_csv(save_file)
    if args.num_classes == 2:
        best_conf = gnn_results.sort_values(by='F1 Score', ascending=False).iloc[0]
    else:
        best_conf = gnn_results.sort_values(by='Weighted F1', ascending=False).iloc[0]
    args.best_num_layers = str(best_conf['NUM_LAYERS'])
    args.best_dropout = str(best_conf['DROPOUT'])
    args.best_norm_type = str(best_conf['NORM_TYPE'])
    return


def run_hovernet(args: Namespace, logger: Logger) -> None:
    """
    Runs the Hovernet inference.

    :param args: The arguments for running Hovernet inference.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Starting hovernet inference.')
    newargs = {
        'nr_types': str(args.num_classes + 1),
        'type_info_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', 'type_info.json'),
        'gpu': args.gpu,
        'nr_inference_workers': '0',
        'model_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', '01', 'net_epoch=50.tar'),
        'batch_size': '10',
        'shape': '518',
        'nr_post_proc_workers': '0',
        'model_mode': 'original',
        'help': False
    }
    newsubargs = {
        'input_dir': args.input_dir,
        'output_dir': os.path.join(args.output_dir, 'tmp_hov'),
        'draw_dot': False,
        'save_qupath': False,
        'save_raw_map': False,
        'mem_usage': '0.2'
    }
    hov_infer(newargs, newsubargs, 'tile')
    return


def run_posthov(args: Namespace, logger: Logger) -> None:
    """
    Performs post-processing steps on Hovernet output.

    :param args: The arguments for running post-processing on Hovernet output.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Parsing Hovernet output')
    logger.info('   From json to geojson.')
    newargs = Namespace(
        json_dir=os.path.join(args.output_dir, 'tmp_hov', 'json'),
        gson_dir=os.path.join(args.output_dir, 'gson_hov'),
        num_classes=args.num_classes
    )
    hovernet2geojson_main(newargs)
    logger.info('   From geojson to pngcsv.')
    newargs = Namespace(
        gson_dir=os.path.join(args.output_dir, 'gson_hov'),
        png_dir=os.path.join(args.output_dir, 'png_hov'),
        csv_dir=os.path.join(args.output_dir, 'csv_hov'),
        num_classes=args.num_classes
    )
    geojson2pngcsv_main(newargs)
    logger.info('   From pngcsv to nodes.csv.')
    create_dir(os.path.join(args.output_dir, 'graphs'))
    newargs = Namespace(
        png_dir=os.path.join(args.output_dir, 'png_hov'),
        orig_dir=args.input_dir,
        output_path=os.path.join(args.output_dir, 'graphs', 'raw'),
        num_workers=args.num_workers
    )
    png2graph_main(newargs)
    logger.info('   Adding hovernet predictions to .nodes.csv.')
    newargs = Namespace(
        json_dir=os.path.join(args.output_dir, 'tmp_hov', 'json'),
        graph_dir=os.path.join(args.output_dir, 'graphs', 'raw'),
        output_dir=os.path.join(args.output_dir, 'graphs', 'hovpreds'),
        num_classes=args.num_classes
    )
    join_hovprob_graph_main(newargs, logger)
    return


def run_graphs(args: Namespace, logger: Logger) -> None:
    """
    Runs the graph inference.

    :param args: The arguments for running graph inference.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Starting graph inference.')
    model_name = 'best_' + args.best_arch + '_' + args.best_num_layers + '_' \
        + args.best_dropout + '_' + args.best_norm_type
    newargs = Namespace(
        node_dir=os.path.join(args.output_dir, 'graphs', 'hovpreds'),
        output_dir=os.path.join(args.output_dir, 'gnn_preds'),
        weights=os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'weights', model_name + '.pth'),
        conf=os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'confs', model_name + '.json'),
        normalizers=os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'normalizers', model_name + '.pkl'),
        num_classes=args.num_classes,
        disable_prior=False,
        disable_morph_feats=False,
    )
    infer_gnn(newargs)
    return


def run_postgraphs(args: Namespace, logger: Logger) -> None:
    """
    Performs post-processing steps on GNN output.

    :param args: The arguments for running post-processing on GNN output.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Parsing gnn output.')
    logger.info('   Converting .nodes.csv to .centroids.csv.')
    newargs = Namespace(
        graph_dir=os.path.join(args.output_dir, 'gnn_preds'),
        centroids_dir=os.path.join(args.output_dir, 'centroids'),
        num_classes=args.num_classes,
    )
    graph2centroids_main(newargs)
    logger.info('   Converting .centroids.csv and .GT_cells.png to .class.csv.')
    newargs = Namespace(
        centroids_dir=os.path.join(args.output_dir, 'centroids'),
        png_dir=os.path.join(args.output_dir, 'png_hov'),
        csv_dir=os.path.join(args.output_dir, 'csv_gnn'),
    )
    centroidspng2csv_main(newargs)
    logger.info('   Converting png/csv to geojson.')
    newargs = Namespace(
        png_dir=os.path.join(args.output_dir, 'png_hov'),
        csv_dir=os.path.join(args.output_dir, 'csv_gnn'),
        gson_dir=os.path.join(args.output_dir, 'gson_gnn'),
        num_classes=args.num_classes
    )
    pngcsv2geojson_main(newargs)
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, help='Internal data and weights directory.', default='./.internals/')
    parser.add_argument('--input-dir', type=str, help='Folder containing patches to process.', required=True)
    parser.add_argument('--output-dir', type=str, help='Folder where to save results. Additional subfolder will be created.', required=True)
    parser.add_argument('--gpu', type=str, default='0')
    parser.add_argument('--num-workers', type=int, default=0)
    parser.add_argument('--best-num-layers', type=str, help='Optimal number of layers when training GNNs.')
    parser.add_argument('--best-dropout', type=str, help='Optimal dropout rate when training GNNs')
    parser.add_argument('--best-norm-type', type=str, help='Optimal type of normalization layers when training GNNs')
    parser.add_argument('--best-arch', type=str, help='Best architecture (convolutional, attention, ...) when training GNNs', required=True, choices=['GCN', 'ATT', 'HATT', 'SAGE', 'BOOST'])
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    create_dir(args.output_dir)

    logger = logging.getLogger('infer_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    if args.best_num_layers is None or args.best_dropout is None or args.best_norm_type is None:
        set_best_configuration(args, logger)
    run_hovernet(args, logger)
    run_posthov(args, logger)
    run_graphs(args, logger)
    run_postgraphs(args, logger)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import numpy as np

from tumourkit.utils.preprocessing import parse_path
from tumourkit.utils.postprocessing import get_N_closest_pairs_dists

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
RSWOOSH_DIR = parse_path(TEST_DIR) + 'rswoosh/'
THRESHOLD = 0.0001

@pytest.mark.parametrize("A,B,result", [
    (
        np.hstack((np.linspace(0,30,100).reshape(-1,1), np.ones((100,1)))),
        np.hstack((np.linspace(0,30,100).reshape(-1,1), np.ones((100,1))*2)), 
        np.ones((10,))
    ),
    (
        np.hstack((np.linspace(0,300,101).reshape(-1,1), np.ones((101,1))))+np.array([[1,0]]),
        np.hstack((np.linspace(0,300,101).reshape(-1,1), np.ones((101,1))*2)), 
        np.ones((20,)) * np.sqrt(2)
    ),
    (
        np.hstack((np.linspace(0,30,11).reshape(-1,1)**2, np.ones((11,1)))),
        np.hstack((np.linspace(0,30,11).reshape(-1,1), np.ones((11,1))*2)), 
        np.array([1, 1, np.sqrt(3**2+1), np.sqrt(3**2+1), np.sqrt(3**2+1), np.sqrt(6**2+1), np.sqrt(6**2+1)])
    )
])
def test_conversion(A,B, result):
    pred = get_N_closest_pairs_dists(A, B, len(result))
    pred = sorted(pred)
    if len(pred) != len(result):
        assert(False)
    for i in range(len(pred)):
        if abs(pred[i] - result[i]) > THRESHOLD:
            assert(False)
    assert(True)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import numpy as np

from tumourkit.preprocessing.geojson2pngcsv import geojson2pngcsv
from tumourkit.preprocessing.pngcsv2geojson import pngcsv2geojson
from tumourkit.utils.preprocessing import get_names, parse_path, read_labels

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PNGCSV_DIR = parse_path(TEST_DIR) + 'pngcsv/'
THRESHOLD = 0.1

def parse_labels(png, csv):
    negatives = set(csv[csv.label==1].id)
    positives = set(csv[csv.label==2].id)
    assert(not (csv.label==0).any())
    def parser(x):
        if x in negatives:
            return 1
        elif x in positives:
            return 2
        else:
            return 0
    vec_parser = np.vectorize(parser)
    return vec_parser(png)

def compare(img1, img2):
    return np.sum(np.abs(img1-img2)) / (np.multiply.reduce(img1.shape))

def check_equal(png1, csv1, png2, csv2):
    img_label1 = parse_labels(png1, csv1)
    img_label2 = parse_labels(png2, csv2)
    assert(len(np.unique(img_label1)) == 3)
    assert(len(np.unique(img_label2)) == 3)
    diff = compare(img_label1, img_label2)
    print(diff)
    return diff < THRESHOLD

@pytest.mark.parametrize("name", get_names(PNGCSV_DIR, '.GT_cells.png'))
def test_conversion(name):
    png_orig, csv_orig = read_labels(name, PNGCSV_DIR, PNGCSV_DIR)
    gson = pngcsv2geojson(png_orig, csv_orig)
    png_new, csv_new = geojson2pngcsv(gson)
    if not check_equal(png_orig,csv_orig, png_new,csv_new):
        print(name)
        assert(False)
    assert(True)



from argparse import Namespace
from logging import Logger
import os
import pandas as pd
from typing import List


class WrongConfigurationError(Exception):
    pass


class HovernetNotFoundError(Exception):
    pass


def check_void(path: str) -> bool:
    """
    Checks if a directory is void (empty) or if a file exists..

    :param path: The path to the directory or file.
    :type path: str

    :return: True if the directory is void (empty) or the file exists, False otherwise.
    :rtype: bool
    """
    if os.path.isdir(path):
        files = os.listdir(path)
        return len(files) == 0
    return not os.path.exists(path)


def check_same_num(dir_list: List[str]) -> bool:
    """
    Checks if all the folders in the list contain the same number of files.

    This function takes a list of folder paths and checks if all the folders contain the same number of files. It counts
    the number of files in each folder and compares them. If the number of files is the same for all folders, the
    function returns True. Otherwise, it returns False.

    :param dir_list: The list of folder paths.
    :type dir_list: List[str]

    :return: True if all folders contain the same number of files, False otherwise.
    :rtype: bool
    """
    num_files = [len([os.listdir(dir) for dir in dir_list])]
    if len(num_files) == 0:
        return True
    num_file = num_files[0]
    for element in num_files:
        if element != num_file:
            return False
    return True


def check_training(args: Namespace, logger: Logger) -> None:
    """
    Checks if the training pipeline was run correctly.

    This function checks if the necessary folders and files from the training pipeline exist. It verifies the presence
    of specific folders and files in the expected locations. If any required folder or file is missing, or if there are
    inconsistencies in the training results, an exception is raised.

    :param args: The command-line arguments.
    :type args: Namespace

    :param logger: The logger object for logging messages.
    :type logger: Logger

    :raises WrongConfigurationError: If any required folder or file is missing, or if there are inconsistencies in the training results.
    """
    logger.info('Checking training pipeline was run correctly.')
    check_list = [
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'confs'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'normalizers'),
        os.path.join(args.root_dir, 'weights', 'classification', 'gnn', 'weights'),
        os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
    ]
    for element in check_list:
        if check_void(element):
            raise WrongConfigurationError(f'The following folder / file was not found: {element}. Try running the training pipeline again.')
    graph_check_list = [
        os.path.join(args.root_dir, 'gnn_logs', 'gnn_results.csv'),
        os.path.join(args.root_dir, 'gnn_logs', 'gcn_results.csv'),
        os.path.join(args.root_dir, 'gnn_logs', 'gat_results.csv'),
    ]
    exist_result_file = False
    graph_result_file = None
    for element in graph_check_list:
        if not check_void(element):
            exist_result_file = True
            graph_result_file = element
    if not exist_result_file:
        graph_folder = os.path.join(args.root_dir, 'gnn_logs')
        raise WrongConfigurationError(f'There is not graph result csv file under {graph_folder}')

    dir_list = [os.path.join(args.root_dir, 'weights', 'classification', 'gnn', dir) for dir in ['confs', 'normalizers', 'weights']]
    if not check_same_num(dir_list):
        raise WrongConfigurationError('Graph weights folders contain different number of files. Try running the training pipeline again')

    results = pd.read_csv(graph_result_file)
    if 'F1 Score' in results.columns and args.num_classes != 2:
        raise WrongConfigurationError('You have trained the GNN models for binary problem but are asking to run research for multiclass.')
    elif 'Micro F1' in results.columns and args.num_classes == 2:
        raise WrongConfigurationError('You have trained the GNN models for multiclass problem but are asking to run research for binary.')



"""
Removes cells with label uncertain from the annotations.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import os
from typing import List, Dict, Any
from tqdm import tqdm
import argparse
from argparse import Namespace
from ..utils.preprocessing import get_names, read_gson, save_geojson


def process_gson(gson: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
    """
    Filters out uncertain classifications from a list of GeoJSON dictionaries.

    Given a list of GeoJSON dictionaries, this function filters out the GeoJSON objects
    that have an uncertain classification. The uncertain classification is determined
    based on the 'name' property under the 'classification' key in the GeoJSON dictionary.

    :param gson: The list of GeoJSON dictionaries.
    :type gson: List[Dict[str, Any]]
    :return: The filtered list of GeoJSON dictionaries without uncertain classifications.
    :rtype: List[Dict[str, Any]]
    """
    return list(filter(lambda x: x['properties']['classification']['name'] != 'uncertain', gson))


def main_with_args(args: Namespace) -> None:
    os.makedirs(args.output_dir, exist_ok=True)
    names = get_names(args.input_dir, '.geojson')
    for name in tqdm(names):
        gson = read_gson(name, args.input_dir)
        out_gson = process_gson(gson)
        save_geojson(out_gson, name, args.output_dir)
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--input-dir', type=str, help='Folder where geojsons are with label uncertain.')
    parser.add_argument('--output-dir', type=str, help='Folder where to save geojsons without label uncertain.')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Module with utility functions referring to finding nearest points in sets.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from scipy.spatial import KDTree
import numpy as np
from typing import List, Tuple, Optional


def generate_tree(centroids: List[Tuple[int, int, int]]) -> KDTree:
    """
    Returns a KDTree object from a list of (x, y, class) tuples.

    :param centroids: A list of (x, y, class) tuples representing the centroids.
    :type centroids: list[tuple(int, int, int)]
    :return: A KDTree object created from the x and y coordinates of the centroids.
    :rtype: KDTree
    """
    centroids_ = np.array(list(map(lambda x: (x[0], x[1]), centroids)))
    return KDTree(centroids_)


def find_nearest(a: Tuple[int, int, int], B: KDTree) -> int:
    """
    Finds the index of the nearest point in a KDTree to a given (x, y, class) tuple.

    :param a: The (x, y, class) tuple to search for the nearest point.
    :type a: tuple(int, int, int)
    :param B: The KDTree to search for the nearest point.
    :type B: KDTree
    :return: The index of the nearest point in the KDTree to the given tuple.
    :rtype: int
    """
    x, y = a[0], a[1]
    dist, idx = B.query([x, y], k=1)
    return idx


def find_nearest_dist_idx(a: Tuple[int, int, int], B: KDTree) -> Tuple[float, int]:
    """
    Finds the distance and index of the nearest point in a KDTree to a given (x, y, class) tuple.

    :param a: The (x, y, class) tuple to search for the nearest point.
    :type a: tuple(int, int, int)
    :param B: The KDTree to search for the nearest point.
    :type B: KDTree
    :return: A tuple containing the distance and index of the nearest point in the KDTree to the given tuple.
    :rtype: tuple(float, int)
    """
    x, y = a[0], a[1]
    dist, idx = B.query([x, y], k=1)
    return dist, idx


Point = Tuple[float, float]
Contour = List[Point]
Cell = Tuple[int, int, Contour]  # id, class


def get_N_closest_pairs_dists(a: Contour, b: Contour, N: int,
                              threshold: Optional[float] = np.inf) -> List[float]:
    """
    Returns the N closest pairs distances between two sets of points.

    :param a: The first set of points as a list of (x, y) tuples.
    :type a: Contour
    :param b: The second set of points as a list of (x, y) tuples.
    :type b: Contour
    :param N: The number of closest pairs to return.
    :type N: int
    :param threshold: An optional threshold to prune the search for distances greater than the threshold.
    :type threshold: float, optional
    :return: A list of the N closest pairs distances between the two sets of points.
    :rtype: List[float]

    The function has a complexity of O((n+m)log(m)), where m is the maximum of the lengths of the two sets of points and n is the minimum.
    """
    if len(a) > len(b):
        tree = KDTree(a)
        query_set = b
    else:
        tree = KDTree(b)
        query_set = a
    cpairs = []
    for point in query_set:
        dist, idx = tree.query(point, k=N, distance_upper_bound=threshold)
        cpairs.extend(dist)
        cpairs.sort()
        cpairs = cpairs[:N]
    return cpairs


def get_N_closest_pairs_idx(a: Contour, b: Contour, N: int,
                            threshold: Optional[float] = np.inf) -> Tuple[List[int], List[int]]:
    """
    Returns the indices of the N closest pairs of points between two sets of points.

    :param a: The first set of points as a list of (x, y) tuples.
    :type a: Contour
    :param b: The second set of points as a list of (x, y) tuples.
    :type b: Contour
    :param N: The number of closest pairs to return.
    :type N: int
    :param threshold: An optional threshold to prune the search for distances greater than the threshold.
    :type threshold: float, optional
    :return: A tuple containing two lists of indices, one for each set of points, of the N closest pairs of points between the two sets.
    :rtype: Tuple(List[int], List[int])

    The function has a complexity of O((n+m)log(m)), where m is the maximum of the lengths of the two sets of points and n is the minimum.
    """
    is_a_tree = len(a) > len(b)
    if is_a_tree:
        tree = KDTree(a)
        query_set = b
    else:
        tree = KDTree(b)
        query_set = a
    cpairs = []
    for idx2, point in enumerate(query_set):
        dist, idx = tree.query(point, k=N, distance_upper_bound=threshold)
        cpairs.extend(zip(dist, idx, [idx2 for _ in range(len(idx))]))
        cpairs.sort()
        cpairs = cpairs[:N]
    dists, idxa, idxb = list(zip(*cpairs))
    if not is_a_tree:
        idxa, idxb = idxb, idxa
    return list(idxa), list(idxb)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import numpy as np

from tumourkit.preprocessing.pngcsv2graph import pngcsv2graph
from tumourkit.utils.preprocessing import get_names, parse_path, read_labels, read_centroids
from tumourkit.utils.nearest import generate_tree, find_nearest_dist_idx

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PNGCSV_DIR = parse_path(TEST_DIR) + 'pngcsv/'
THRESHOLD = 10


@pytest.mark.parametrize("name", get_names(PNGCSV_DIR, '.GT_cells.png')[-5:])
def test_pngcsv2graph(name):
    png, csv = read_labels(name, PNGCSV_DIR, PNGCSV_DIR)
    false_img = np.ones((*png.shape, 3)) * png.reshape((*png.shape, 1))
    false_img = false_img.astype(np.uint8)
    graph = pngcsv2graph(false_img, png, csv)
    graph.set_index('id', inplace=True)
    graph.sort_index(inplace=True)

    centroids = read_centroids(name, PNGCSV_DIR)
    centr_tree = generate_tree(centroids)
    for _, row in graph.iterrows():
        dist, idx = find_nearest_dist_idx(row[['X','Y', 'class']], centr_tree)
        closest_cls = centroids[idx][2]
        assert(dist < THRESHOLD and closest_cls == row['class'])



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import tumourkit as tt

project = 'tumourkit'
copyright = '2023, Jose Pérez Cano'
author = 'Jose Pérez Cano'
release = tt.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.duration',
    'sphinx_rtd_theme',
    'sphinx.ext.doctest',
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
]
autosummary_generate = True

templates_path = ['_templates']
exclude_patterns = ['**/__init__.py']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ['_static']


"""
Module with utility functions for preprocessing.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import geojson
import os
from geojson import Feature, Polygon
import cv2
import pandas as pd
import numpy as np
import numpy.ma as ma
import json
from typing import Dict, Any, List, Tuple, Optional


def parse_path(path: str) -> str:
    """
    Parses a file path and checks for a trailing slash.

    :param path: The file path to parse.
    :type path: str
    :return: The parsed file path with a trailing slash.
    :rtype: str
    """
    if path[-1] != '/':
        return path + '/'
    return path


def create_dir(path: str) -> None:
    """
    Creates a directory at the specified path if it does not already exist.

    :param path: The path of the directory to create.
    :type path: str
    """
    if not os.path.isdir(path):
        os.mkdir(path)


def get_names(path: str, pattern: str) -> List[str]:
    """
    This function returns a list of all files in a directory located at <path> that contain the substring <pattern> in their name. The <pattern> substring is then removed from the names and the resulting list is returned.

    :param path: The path of the directory to search in.
    :type path: str
    :param pattern: The substring to search for in file names.
    :type pattern: str
    :return: A list of file names without the <pattern> substring.
    :rtype: List[str]
    """
    names_raw = os.listdir(path)
    names = []
    for name in names_raw:
        if pattern in name:
            names.append(name[:-len(pattern)])
    return names


def read_names(file_path: str) -> List[str]:
    """
    Returns a list of file names in the specified path that contain the given pattern.

    :param path: The path to search for files.
    :type path: str
    :param pattern: The pattern to search for in file names.
    :type pattern: str
    :return: A list of file names in the specified path that contain the given pattern.
    :rtype: List[str]
    """
    with open(file_path, 'r') as f:
        lines = f.readlines()
        files = [line.strip() for line in lines]
    return files


def read_png(name: str, png_dir: str) -> np.ndarray:
    """
    Reads a PNG file from the specified directory and returns its image data as a NumPy array.

    :param name: The name of the PNG file (without extension).
    :type name: str
    :param png_dir: The directory containing the PNG file.
    :type png_dir: str
    :return: A NumPy array containing the image data.
    :rtype: np.ndarray or None

    This function reads a PNG file from the specified directory and returns its image data as a NumPy array.
    The function expects the PNG file to have a specific filename format: '<name>.GT_cells.png'.
    If the function is unable to read the PNG file or encounters an error, it returns `None`.
    """
    try:
        img = cv2.imread(os.path.join(png_dir, name + '.GT_cells.png'), -1)
        return img
    except Exception:
        return None


def read_csv(name: str, csv_dir: str) -> Tuple[np.ndarray, pd.DataFrame]:
    """
    Reads a CSV file from the specified directory and returns its contents as a Pandas DataFrame.

    :param name: The name of the CSV file (without extension).
    :type name: str
    :param csv_dir: The directory containing the CSV file.
    :type csv_dir: str
    :return: A tuple containing the ID and label columns of the CSV file as NumPy arrays, and the entire DataFrame.
    :rtype: Tuple[np.ndarray, np.ndarray, pd.DataFrame] or None

    This function reads a CSV file from the specified directory and returns its contents as a Pandas DataFrame.
    The function expects the CSV file to have a specific filename format: '<name>.class.csv'.
    If the function is unable to read the CSV file or encounters an error, it returns `None`.
    """
    try:
        csv = pd.read_csv(os.path.join(csv_dir, name + '.class.csv'), header=None)
        csv.columns = ['id', 'label']
        return csv
    except Exception:
        return None


def read_gson(name: str, path: str) -> List[Dict[str, Any]]:
    """
    Reads the GeoJSON file at the specified path with the given name.

    :param name: The base name of the file (without extension).
    :type name: str
    :param path: The path to the directory containing the file.
    :type path: str
    :return: The contents of the GeoJSON file as a list of dictionaries.
    :rtype: List[Dict[str, Any]]
    """
    with open(os.path.join(path, name + '.geojson'), 'r') as f:
        gson = geojson.load(f)
    return gson


def read_labels(name: str, png_dir: str, csv_dir: str) -> Tuple[np.ndarray, pd.DataFrame]:
    """
    Reads a PNG and CSV file from the specified directories and returns their contents as a tuple.

    :param name: The base name of the files (without extensions).
    :type name: str
    :param png_dir: The directory containing the PNG file.
    :type png_dir: str
    :param csv_dir: The directory containing the CSV file.
    :type csv_dir: str
    :return: A tuple containing the NumPy array of the PNG file and the Pandas DataFrame of the CSV file.
    :rtype: Tuple[np.ndarray, pd.DataFrame] or None

    This function reads a PNG and CSV file from the specified directories and returns their contents as a tuple.
    The function expects the PNG and CSV files to have specific filename formats: '<name>.GT_cells.png' and '<name>.class.csv'.
    If the function is unable to read either file or encounters an error, it returns (`None`,`None`).
    """
    try:
        img = read_png(name, png_dir)
        csv = read_csv(name, csv_dir)
        return img, csv
    except Exception:
        return None, None


def read_json(json_path: str) -> Dict[str, Any]:
    """
    Reads a Hovernet JSON file from the specified path and returns the nuclei information as a dictionary.

    :param json_path: The path to the Hovernet JSON file.
    :type json_path: str
    :return: A dictionary containing the nuclei information.
    :rtype: Dict[str, Any] or None

    This function reads a Hovernet JSON file from the specified path and returns the nuclei information as a dictionary.
    If the function is unable to read the file or encounters an error, it returns `None`.
    The function expects the JSON file to contain a 'nuc' key with the nuclei information.
    """
    with open(json_path) as json_file:
        data = json.load(json_file)
        nuc_info = data['nuc']
    return nuc_info


def read_centroids(name: str, path: str) -> np.ndarray:
    """
    Reads a CSV file from the specified directory containing centroids and returns their contents as a NumPy array.

    :param name: The base name of the CSV file (without extension).
    :type name: str
    :param path: The directory containing the CSV file.
    :type path: str
    :return: A NumPy array containing the centroids with class -1 removed and coordinates converted to integers.
    :rtype: np.ndarray

    This function reads a CSV file from the specified directory containing centroids and returns their contents as a NumPy array.
    The function expects the CSV file to have specific column names: 'X', 'Y', and 'class'.
    Centroids with class -1 are dropped and the coordinates are converted to integers.
    """
    centroid_csv = pd.read_csv(os.path.join(path, name + '.centroids.csv'))
    centroid_csv = centroid_csv.drop(centroid_csv[centroid_csv['class'] == -1].index)
    return centroid_csv.to_numpy(dtype=int)


def read_graph(name: str, graph_dir: str) -> pd.DataFrame:
    """
    Reads a graph in CSV format from the specified directory and returns it as a Pandas DataFrame.

    :param name: The base name of the graph file (without extension).
    :type name: str
    :param graph_dir: The directory containing the graph file.
    :type graph_dir: str
    :return: The graph nodes as a Pandas DataFrame.
    :rtype: pd.DataFrame or None

    This function reads a graph in CSV format from the specified directory and returns it as a Pandas DataFrame.
    The function expects the graph file to have a specific filename format: '<name>.nodes.csv'.
    If the function is unable to read the file or encounters an error, it returns `None`.
    """
    df = pd.read_csv(os.path.join(graph_dir, name + '.nodes.csv'))
    return df


Point = Tuple[float, float]
Contour = List[Point]


def format_contour(contour: Contour) -> Contour:
    """
    Formats a contour in cv2.findContours format to an array of shape (N,2).
    Additionally, the first point is added to the end to close the contour.

    :param contour: The contour to be formatted.
    :type contour: Contour
    :return: The formatted contour.
    :rtype: Contour
    """
    new_contour = np.reshape(contour, (-1, 2)).tolist()
    new_contour.append(new_contour[0])
    return new_contour


def create_geojson(contours: List[Tuple[int, int]], num_classes: Optional[int] = 2) -> List[Dict[str, Any]]:
    """
    Converts a list of contours and their labels to a list of dictionaries
    containing GeoJSON-formatted data.

    :param contours: A list of pairs (contour, label) where contour is a list of points starting
        and finishing in the same point and label is an integer representing the class of the
        cell (1: non-tumour, 2: tumour).
    :type contours: List[Tuple[int, int]]
    :param num_classes: The number of classes in the data. Defaults to 2.
    :type num_classes: Optional[int]
    :return: A list of dictionaries with the GeoJSON format of QuPath.
    :rtype: List[Dict[str, Any]]

    This function converts a list of contours and their labels to a list of dictionaries
    containing GeoJSON-formatted data. The function expects the contours to be a list of
    pairs (contour, label), where contour is a list of points starting and finishing in the
    same point, and label is an integer representing the class of the cell (1: non-tumour,
    2: tumour).

    The function returns a list of dictionaries with the GeoJSON format of QuPath. The
    number of classes can be specified with the num_classes parameter. By default, it is set
    to 2, but it can be set to any integer greater than or equal to 1.

    The function uses a list of labels and colour codes to create the classification data for
    the output dictionaries. The labels and colour codes are stored in the label_dict and
    colour_dict lists, respectively. The label_dict list contains the names of the labels and
    the colour_dict list contains the corresponding RGB colour codes for each label. By default,
    the label_dict list contains the values ["background", "non-tumour", "tumour", "segmented"]
    and the colour_dict list contains the value [-9408287, -9408287, -9408287, -9408287]. If
    the num_classes parameter is set to a value other than 2, the label_dict list is updated
    to contain the values ["background", "Class1", "Class2", ..., "ClassN"], where N is the
    number of classes, and the colour_dict list is updated to contain the value
    [-9408287, ..., -9408287] where the length of the list is N+1. Although this list is
    superfluous and not used by QuPath.

    Each dictionary in the output list represents a cell contour and contains a Polygon
    object with the contour points, a classification object with the name and colour of the
    label, and a boolean indicating whether the object is locked. The dictionary is in the
    format expected by QuPath for reading annotations.
    """
    label_dict = ["background", "non-tumour", "tumour", "segmented"]
    colour_dict = [-9408287, -9408287, -9408287, -9408287]
    if num_classes != 2:
        label_dict = ["background"] + ["Class" + str(i) for i in range(1, num_classes + 1)]
    colour_dict = [-9408287] * (num_classes + 1)
    features = []
    for contour, label in contours:
        assert (label > 0)
        points = Polygon([contour])
        properties = {
            "object_type": "annotation",
            "classification": {
                "name": label_dict[label],
                "colorRGB": colour_dict[label]
            },
            "isLocked": False
        }
        feat = Feature(geometry=points, properties=properties)
        features.append(feat)
    return features


def save_png(png: np.ndarray, png_path: str, name: str) -> None:
    """
    Saves a PNG image to a file.

    :param png: The image to save.
    :type png: np.ndarray
    :param png_path: The path to the folder where the PNG file should be saved.
    :type png_path: str
    :param name: The name of the PNG file (without extension).
    :type name: str

    The file extension is set to '.GT_cells.png'.
    """
    png = np.array(png, dtype=np.uint16)
    cv2.imwrite(os.path.join(png_path, name + '.GT_cells.png'), png)


def save_csv(csv: pd.DataFrame, csv_path: str, name: str) -> None:
    """
    Saves the given CSV file to the specified directory with the specified name and a '.class.csv' extension.

    :param csv: The CSV file to save.
    :type csv: pd.DataFrame
    :param csv_path: The directory to save the CSV file in.
    :type csv_path: str
    :param name: The base name for the file (without extensions).
    :type name: str
    """
    csv.to_csv(os.path.join(csv_path, name + '.class.csv'), index=False, header=False)


def save_pngcsv(png: np.ndarray, csv: pd.DataFrame, png_path: str, csv_path: str, name: str) -> None:
    """
    Saves the given PNG and CSV files to the specified directories with the specified name.

    :param png: The PNG file to save.
    :type png: np.ndarray
    :param csv: The CSV file to save.
    :type csv: pd.DataFrame
    :param png_path: The directory to save the PNG file in.
    :type png_path: str
    :param csv_path: The directory to save the CSV file in.
    :type csv_path: str
    :param name: The base name for the files (without extensions).
    :type name: str

    The PNG file is saved with a '.GT_cells.png' extension, and the CSV file is saved with a '.class.csv' extension.
    """
    save_png(png, png_path, name)
    save_csv(csv, csv_path, name)


def save_centroids(centroids: np.ndarray, centroids_dir: str, name: str) -> None:
    """
    Saves the given centroids to the specified directory with the specified name and a '.centroids.csv' extension.

    :param centroids: The centroids to save.
    :type centroids: np.ndarray
    :param centroids_dir: The directory to save the centroids in.
    :type centroids_dir: str
    :param name: The base name for the file (without extensions).
    :type name: str
    """
    df = pd.DataFrame(centroids, columns=['X', 'Y', 'class'])
    df.to_csv(os.path.join(centroids_dir, name + '.centroids.csv'), index=False)


def save_graph(graph: pd.DataFrame, path: str) -> None:
    """
    Saves the given graph as a CSV file to the specified path.

    :param graph: The graph to save.
    :type graph: pd.DataFrame
    :param path: The path to save the graph file to.
    :type path: str
    """
    graph.set_index('id', inplace=True)
    graph.sort_index(inplace=True)
    graph.to_csv(path)


def save_geojson(gson: List[Dict[str, Any]], name: str, path: str) -> None:
    """
    Save a list of geojson features to a file with the given name at the
    specified path.

    :param gson: A list of geojson features to save.
    :type gson: List[Dict[str, Any]]
    :param name: The name of the file to save.
    :type name: str
    :param path: The path to save the file at.
    :type path: str
    """
    with open(os.path.join(path, name + '.geojson'), 'w') as f:
        geojson.dump(gson, f, sort_keys=True, indent=2)


def get_centroid_by_id(img: np.ndarray, idx: int) -> Tuple[int, int]:
    """
    Given an image and an id representing a component, returns the centroid
    of the component as a tuple (x, y).

    :param img: A NumPy array representing the image.
    :type img: np.ndarray
    :param idx: An integer representing the id of the component.
    :type idx: int
    :return: A tuple representing the centroid of the component.
    :rtype: Tuple[int, int]

    If there are no pixels with the given id in the image, it returns (-1, -1).
    """
    X, Y = np.where(img == idx)
    if len(X) == 0 or len(Y) == 0:
        return -1, -1
    return X.mean(), Y.mean()


def get_mask(png: np.ndarray, idx: int) -> np.ndarray:
    """
    Given segmentation mask with indices as pixel values,
    returns the mask corresponding to the given index.

    :param png: Segmentation mask with indices as pixel values.
    :type png: np.ndarray
    :param idx: Index of the mask to retrieve.
    :type idx: int
    :return: The mask corresponding to the given index.
    :rtype: np.ndarray
    """
    png_aux = png.copy()
    png_aux[png_aux != idx] = 0
    png_aux[png_aux != 0] = 1
    return png_aux


def apply_mask(img: np.ndarray, mask: np.ndarray) -> Tuple[np.ndarray, int, int]:
    """
    Apply a binary mask to an RGB image.

    :param img: RGB image.
    :type img: np.ndarray
    :param mask: Binary mask.
    :type mask: np.ndarray
    :return: Tuple containing the masked image, the bounded box mask, the center of the mask in the masked image.
    :rtype: Tuple[np.ndarray, np.ndarray, int, int]

    For efficiency only the bounded box of the mask is returned, together with the center of the box.
    Coordinates are indices of array.
    """
    img_aux = img.copy()
    img_aux = img_aux * mask.reshape(*mask.shape, 1)
    x, y, w, h = cv2.boundingRect((mask * 255).astype(np.uint8))
    X, Y = np.where(mask == 1)
    if len(X) == 0 or len(Y) == 0:
        cx, cy = -1, -1
    else:
        cx, cy = X.mean(), Y.mean()
    return img_aux[y:y + h, x:x + w].copy(), mask[y:y + h, x:x + w].copy(), cx, cy


def read_image(name: str, path: str) -> np.array:
    """
    Given name image (without extension) and folder path,
    returns array with pixel values (RGB).

    :param name: Name of the image (without extension).
    :type name: str
    :param path: Path to the folder where the image is located.
    :type path: str
    :return: Array with pixel values (RGB).
    :rtype: np.array
    """
    aux = cv2.imread(os.path.join(path, name + '.png'))
    return cv2.cvtColor(aux, cv2.COLOR_BGR2RGB)


def compute_perimeter(c: Contour) -> float:
    """
    Compute the perimeter of a given contour.
    Prerequisites: First and last point must be the same.

    :param c: Contour represented as a list of points (x,y).
              The first and last point must be the same.
    :type c: List[Tuple[float, float]]
    :return: The perimeter of the contour.
    :rtype: float
    """
    diff = np.diff(c, axis=0)
    dists = np.hypot(diff[:, 0], diff[:, 1])
    return dists.sum()


def extract_features(msk_img: np.ndarray, bin_msk: np.ndarray, debug=False) -> Dict[str, np.ndarray]:
    """
    Extracts features from a given RGB bounding box of a cell and its mask.

    :param msk_img: RGB bounding box of a cell.
    :type msk_img: np.ndarray
    :param bin_msk: Binary mask of the cell.
    :type bin_msk: np.ndarray
    :param debug: Flag indicating if debug mode is enabled.
    :type debug: bool
    :return: A dictionary containing different extracted features.
    :rtype: Dict[str, np.ndarray]
    """
    if len(msk_img) == 0 or msk_img.max() == 0:
        return {}
    gray_msk = cv2.cvtColor(msk_img, cv2.COLOR_RGB2GRAY)
    # bin_msk = (gray_msk > 0) * 1
    feats = {}
    feats['area'] = bin_msk.sum()

    if debug:
        import pdb
        pdb.set_trace()
    contours, _ = cv2.findContours(
        (bin_msk * 255).astype(np.uint8),
        mode=cv2.RETR_TREE, method=cv2.CHAIN_APPROX_SIMPLE
    )
    contour = format_contour(contours[0])
    feats['perimeter'] = compute_perimeter(contour)

    gray_msk = ma.masked_array(
        gray_msk, mask=1 - bin_msk
    )
    feats['std'] = gray_msk.std()

    msk_img = ma.masked_array(
        msk_img, mask=1 - np.repeat(bin_msk.reshape((*bin_msk.shape, 1)), 3, axis=2)
    )
    red = msk_img[:, :, 0].compressed()
    red_bins, _ = np.histogram(red, bins=5, density=True)
    feats['red'] = red_bins
    green = msk_img[:, :, 1].compressed()
    green_bins, _ = np.histogram(green, bins=5, density=True)
    feats['green'] = green_bins
    blue = msk_img[:, :, 2].compressed()
    blue_bins, _ = np.histogram(blue, bins=5, density=True)
    feats['blue'] = blue_bins
    return feats


def add_node(graph: Dict[str, float], feats: Dict[str, np.ndarray]) -> None:
    """
    Given a dictionary with vectorial features,
    converts them into one column per dimension
    and add them into the global dictionary.

    :param graph: Dictionary with graph node data.
    :type graph: Dict[str, float]
    :param feats: Dictionary with vectorial features of a graph node.
    :type feats: Dict[str, np.ndarray]
    :return: None. The graph dictionary is modified in place.
    """
    for k, v in feats.items():
        if hasattr(v, '__len__'):
            for i, coord in enumerate(v):
                if k + str(i) not in graph:
                    graph[k + str(i)] = []
                graph[k + str(i)].append(coord)
        else:
            if k not in graph:
                graph[k] = []
            graph[k].append(v)



"""
Script to train and save several GNN configurations.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import math
from typing import Optional, Tuple, Dict, List, Any
from sklearn.preprocessing import Normalizer
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim import Optimizer
import dgl
from dgl.dataloading import GraphDataLoader
from .read_graph import GraphDataset
from .models.gcn import GCN
from .models.hgao import HardGAT
from .models.gat import GAT
import argparse
from argparse import Namespace
from torch.utils.tensorboard import SummaryWriter
import warnings
import os
from concurrent.futures import ThreadPoolExecutor
import json
import pickle
from ..utils.preprocessing import parse_path, create_dir
from ..utils.classification import metrics_from_predictions
warnings.filterwarnings('ignore')


def evaluate(
        loader: GraphDataLoader,
        model: nn.Module,
        device: str,
        writer: Optional[SummaryWriter] = None,
        epoch: Optional[int] = None,
        log_suffix: Optional[str] = None,
        num_classes: Optional[str] = 2,
        ) -> List[float]:
    """
    Evaluates model in loader.
    Logs to tensorboard with suffix log_suffix.
    Returns the model in evaluation mode.
    """
    model.eval()
    preds, labels, probs = np.array([]).reshape(0, 1), np.array([]).reshape(0, 1), np.array([]).reshape(0, 1 if num_classes == 2 else num_classes)
    for g in loader:
        g = g.to(device)
        # self-loops
        g = dgl.remove_self_loop(g)
        g = dgl.add_self_loop(g)
        # data
        features = g.ndata['X']
        # Forward
        logits = model(g, features)
        pred = logits.argmax(1).detach().cpu().numpy().reshape(-1, 1)
        preds = np.vstack((preds, pred))
        if num_classes == 2:
            prob = F.softmax(logits, dim=1).detach().cpu().numpy()[:, 1].reshape(-1, 1)
        else:
            prob = F.softmax(logits, dim=1).detach().cpu().numpy()
        probs = np.vstack((probs, prob))
        label = g.ndata['y'].detach().cpu().numpy().reshape(-1, 1)
        labels = np.vstack((labels, label))
    # Compute metrics on validation
    if num_classes == 2:
        acc, f1, auc, perc_err, ece = metrics_from_predictions(labels, preds, probs, 2)
        # Tensorboard
        if writer is not None:
            assert (log_suffix is not None and epoch is not None)
            writer.add_scalar('Accuracy/' + log_suffix, acc, epoch)
            writer.add_scalar('F1/' + log_suffix, f1, epoch)
            writer.add_scalar('ROC_AUC/' + log_suffix, auc, epoch)
            writer.add_scalar('Percentage Error/' + log_suffix, perc_err, epoch)
            writer.add_scalar('ECE/' + log_suffix, ece, epoch)
        return f1, acc, auc, perc_err, ece
    else:
        micro, macro, weighted, ece = metrics_from_predictions(labels, preds, probs, num_classes)
        # Tensorboard
        if writer is not None:
            writer.add_scalar('Accuracy/' + log_suffix, micro, epoch)
            writer.add_scalar('Macro F1/' + log_suffix, macro, epoch)
            writer.add_scalar('Weighted F1/' + log_suffix, weighted, epoch)
            writer.add_scalar('ECE/' + log_suffix, ece, epoch)
        return micro, macro, weighted, ece


def train_one_iter(
        tr_loader: GraphDataLoader,
        model: nn.Module,
        device: str,
        optimizer: Optimizer,
        epoch: int,
        writer: SummaryWriter,
        num_classes: int,
        ) -> None:
    """
    Trains for one iteration, as the name says.
    """
    model.train()
    for step, tr_g in enumerate(tr_loader):
        tr_g = tr_g.to(device)
        # self-loops
        tr_g = dgl.remove_self_loop(tr_g)
        tr_g = dgl.add_self_loop(tr_g)
        # data
        features = tr_g.ndata['X']
        labels = tr_g.ndata['y']
        # Forward
        logits = model(tr_g, features)
        loss = F.cross_entropy(logits, labels)
        # Backward
        optimizer.zero_grad()
        loss.backward()
        optimizer.step()
        # Compute metrics on training
        preds = logits.argmax(1).detach().cpu().numpy()
        labels = labels.detach().cpu().numpy()
        if num_classes == 2:
            probs = F.softmax(logits, dim=1).detach().cpu().numpy()[:, 1]
            train_acc, train_f1, train_auc, train_perc_err, train_ece = metrics_from_predictions(labels, preds, probs, 2)
            # Tensorboard
            writer.add_scalar('Accuracy/train', train_acc, step + len(tr_loader) * epoch)
            writer.add_scalar('F1/train', train_f1, step + len(tr_loader) * epoch)
            writer.add_scalar('ROC_AUC/train', train_auc, step + len(tr_loader) * epoch)
            writer.add_scalar('ECE/train', train_ece, step + len(tr_loader) * epoch)
            writer.add_scalar('Percentage Error/train', train_perc_err, step + len(tr_loader) * epoch)
        else:
            probs = F.softmax(logits, dim=1).detach().cpu().numpy()
            train_micro, train_macro, train_weighted, train_ece = metrics_from_predictions(labels, preds, probs, num_classes)
            # Tensorboard
            writer.add_scalar('Accuracy/train', train_micro, step + len(tr_loader) * epoch)
            writer.add_scalar('Macro F1/train', train_macro, step + len(tr_loader) * epoch)
            writer.add_scalar('Weighted F1/train', train_weighted, step + len(tr_loader) * epoch)
            writer.add_scalar('ECE/train', train_ece, step + len(tr_loader) * epoch)


def train(
        save_dir: str,
        save_weights: bool,
        tr_loader: GraphDataLoader,
        val_loader: GraphDataLoader,
        model: nn.Module,
        optimizer: Optimizer,
        writer: SummaryWriter,
        n_early: int,
        device: Optional[str] = 'cpu',
        check_iters: Optional[int] = -1,
        conf: Optional[Dict[str, Any]] = None,
        normalizers: Optional[Tuple[Normalizer]] = None,
        num_classes: Optional[int] = 2,
        ) -> None:
    """
    Train the model with early stopping on F1 score (weighted) or until 1000 iterations.
    """
    model = model.to(device)
    n_epochs = 1000
    best_val_f1 = 0
    early_stop_rounds = 0
    for epoch in range(n_epochs):
        train_one_iter(tr_loader, model, device, optimizer, epoch, writer, num_classes)
        val_metrics = evaluate(val_loader, model, device, writer, epoch, 'validation', num_classes=num_classes)
        if num_classes == 2:
            val_f1, val_acc, val_auc, val_perc_error, val_ece = val_metrics
        else:
            val_micro, val_macro, val_f1, val_ece = val_metrics
        # Save checkpoint
        if save_weights and check_iters != -1 and epoch % check_iters == 0:
            save_model(save_dir, model, conf, normalizers, prefix='last_')
        # Early stopping
        if val_f1 > best_val_f1:
            best_val_f1 = val_f1
            early_stop_rounds = 0
            if save_weights:
                save_model(save_dir, model, conf, normalizers, prefix='best_')
        elif early_stop_rounds < n_early:
            early_stop_rounds += 1
        else:
            return


def load_dataset(
        train_node_dir: str,
        val_node_dir: str,
        test_node_dir: str,
        bsize: int,
        remove_prior: Optional[bool] = False,
        remove_morph: Optional[bool] = False,
        ) -> Tuple[GraphDataLoader, GraphDataLoader, GraphDataLoader]:
    """
    Creates Torch dataloaders for training.
    Folder structure:
    node_dir:
     - train
      - graphs
       - file1.nodes.csv
       ...
     - validation
      - graphs
       - file1.nodes.csv
       ...
     - test
      - graphs
       - file1.nodes.csv
       ...
    """
    train_dataset = GraphDataset(
        node_dir=train_node_dir, remove_morph=remove_morph, remove_prior=remove_prior,
        max_dist=200, max_degree=10, column_normalize=True)
    train_dataloader = GraphDataLoader(train_dataset, batch_size=bsize, shuffle=True)
    val_dataset = GraphDataset(
        node_dir=val_node_dir, remove_morph=remove_morph, remove_prior=remove_prior,
        max_dist=200, max_degree=10, normalizers=train_dataset.get_normalizers())
    val_dataloader = GraphDataLoader(val_dataset, batch_size=1, shuffle=False)
    test_dataset = GraphDataset(
        node_dir=test_node_dir, remove_morph=remove_morph, remove_prior=remove_prior,
        max_dist=200, max_degree=10, normalizers=train_dataset.get_normalizers())
    test_dataloader = GraphDataLoader(test_dataset, batch_size=1, shuffle=False)
    return train_dataloader, val_dataloader, test_dataloader


def generate_configurations(max_confs: int, model_name: str) -> List[Dict[str, int]]:
    """
    Generates a grid in the search space with no more than max_confs configurations.
    Parameters changed: NUM_LAYERS, DROPOUT, NORM_TYPE
    """
    num_layers_confs = int(math.sqrt(max_confs / 2))
    num_dropout_confs = int(max_confs // (2 * num_layers_confs))
    assert (2 * num_layers_confs * num_dropout_confs <= max_confs)
    assert num_layers_confs <= 15, 'Too many layers'
    confs = []
    for num_layers in np.linspace(1, 15, num_layers_confs):
        num_layers = int(num_layers)
        for dropout in np.linspace(0, 0.9, num_dropout_confs):
            conf = {}
            conf['MODEL_NAME'] = model_name
            conf['NUM_LAYERS'] = num_layers
            conf['DROPOUT'] = dropout
            conf['NORM_TYPE'] = 'bn'
            confs.append(conf)

            conf = {}
            conf['MODEL_NAME'] = model_name
            conf['NUM_LAYERS'] = num_layers
            conf['DROPOUT'] = dropout
            conf['NORM_TYPE'] = None
            confs.append(conf)
    return confs


def load_model(conf: Dict[str, Any], num_classes: int, num_feats: int) -> nn.Module:
    """
    Available models: GCN, ATT, HATT, SAGE, BOOST
    Configuration space: NUM_LAYERS, DROPOUT, NORM_TYPE
    """
    hidden_feats = 100
    if conf['MODEL_NAME'] == 'GCN':
        return GCN(num_feats, hidden_feats, num_classes, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'])
    if conf['MODEL_NAME'] == 'ATT' or conf['MODEL_NAME'] == 'HATT':
        num_heads = 8
        num_out_heads = 1
        heads = ([num_heads] * conf['NUM_LAYERS']) + [num_out_heads]
        if conf['MODEL_NAME'] == 'ATT':
            return GAT(num_feats, hidden_feats, num_classes, heads, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'])
        return HardGAT(num_feats, hidden_feats, num_classes, heads, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'])
    assert False, 'Model not implemented.'


def create_results_file(filename: str, num_classes: int) -> None:
    """
    Creates header of .csv result file to append results.
    filename must not contain extension.
    """
    if num_classes == 2:
        with open(filename + '.csv', 'w') as f:
            print('F1 Score,Accuracy,ROC AUC,PERC ERR,ECE,NUM_LAYERS,DROPOUT,NORM_TYPE', file=f)
    else:
        with open(filename + '.csv', 'w') as f:
            print('Micro F1,Macro F1,Weighted F1,ECE,NUM_LAYERS,DROPOUT,NORM_TYPE', file=f)


def append_results(
        filename: str,
        f1: float, acc: float, auc: float,
        num_layers: int, dropout: float, bn_type: str,
        ece: float, perc_err: Optional[float] = None
        ) -> None:
    """
    Appends result to given filename.
    filename must not contain extension.
    """
    with open(filename + '.csv', 'a') as f:
        if perc_err is not None:
            print(f1, acc, auc, perc_err, ece, num_layers, dropout, bn_type, file=f, sep=',')
        else:
            print(f1, acc, auc, ece, num_layers, dropout, bn_type, file=f, sep=',')


def name_from_conf(conf: Dict[str, Any]) -> str:
    """
    Generates a name from the configuration object.
    """
    return conf['MODEL_NAME'] + '_' + str(conf['NUM_LAYERS']) + '_' \
        + str(conf['DROPOUT']) + '_' + str(conf['NORM_TYPE'])


def save_model(
        save_dir: str,
        model: nn.Module,
        conf: Dict[str, Any],
        normalizers: Tuple[Normalizer],
        prefix: Optional[str] = ''
        ) -> None:
    """
    Save model weights and configuration file to SAVE_DIR
    """
    name = prefix + name_from_conf(conf)
    state_dict = model.state_dict()
    torch.save(state_dict, os.path.join(save_dir, 'weights', name + '.pth'))
    with open(os.path.join(save_dir, 'confs', name + '.json'), 'w') as f:
        json.dump(conf, f)
    with open(os.path.join(save_dir, 'normalizers', name + '.pkl'), 'wb') as f:
        pickle.dump(normalizers, f)


def train_one_conf(
        args: Namespace,
        conf: Dict[str, Any],
        train_dataloader: GraphDataLoader,
        val_dataloader: GraphDataLoader,
        test_dataloader: GraphDataLoader,
        log_dir: str,
        save_weights: bool,
        save_dir: str,
        num_classes: int,
        num_feats: int
        ) -> Tuple[List[float], nn.Module, Dict[str, Any]]:
    # Tensorboard logs
    writer = SummaryWriter(log_dir=os.path.join(log_dir, name_from_conf(conf)))
    # Model
    model = load_model(conf, num_classes, num_feats)
    optimizer = torch.optim.Adam(model.parameters(), lr=0.001)
    # Train
    train(
        save_dir, save_weights, train_dataloader, val_dataloader,
        model, optimizer, writer, args.early_stopping_rounds,
        args.device, args.checkpoint_iters, conf, train_dataloader.dataset.get_normalizers(),
        num_classes=num_classes
    )
    test_metrics = evaluate(test_dataloader, model, args.device, num_classes=num_classes)
    model = model.cpu()
    if num_classes == 2:
        test_f1, test_acc, test_auc, test_perc_err, test_ece = test_metrics
        return test_f1, test_acc, test_auc, test_perc_err, test_ece, model, conf
    else:
        test_micro, test_macro, test_weighted, test_ece = test_metrics
        return test_micro, test_macro, test_weighted, test_ece, model, conf


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--train-node-dir', type=str, required=True,
                        help='Path to folder containing train folder with .nodes.csv files.')
    parser.add_argument('--validation-node-dir', type=str, required=True,
                        help='Path to folder containing validation folder with .nodes.csv files.')
    parser.add_argument('--test-node-dir', type=str, required=True,
                        help='Path to folder containing test folder with .nodes.csv files.')
    parser.add_argument('--log-dir', type=str, required=True,
                        help='Path to save tensorboard logs.')
    parser.add_argument('--early-stopping-rounds', type=int, required=True,
                        help='Number of epochs needed to consider convergence when worsening.')
    parser.add_argument('--batch-size', type=int, required=True,
                        help='Batch size. No default.')
    parser.add_argument('--model-name', type=str, required=True, choices=['GCN', 'ATT', 'HATT', 'SAGE', 'BOOST'],
                        help='Which model to use. Options: GCN, ATT, HATT, SAGE, BOOST')
    parser.add_argument('--save-file', type=str, required=True,
                        help='Name to file where to save the results. Must not contain extension.')
    parser.add_argument('--num-confs', type=int, default=50,
                        help='Upper bound on the number of configurations to try.')
    parser.add_argument('--save-dir', type=str,
                        help='Folder to save models weights and confs.')
    parser.add_argument('--device', type=str, choices=['cpu', 'cuda'], default='cpu',
                        help='Device to execute. Either cpu or cuda. Default: cpu.')
    parser.add_argument('--num-workers', type=int, default=1,
                        help='Number of processors to use. Default: 1.')
    parser.add_argument('--checkpoint-iters', type=int, default=-1,
                        help='Number of iterations at which to save model periodically while training. Set to -1 for no checkpointing. Default: -1.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    parser.add_argument('--disable-prior', action='store_true', help='If True, remove hovernet probabilities from node features.')
    parser.add_argument('--disable-morph-feats', action='store_true', help='If True, remove morphological features from node features.')
    return parser


def main_with_args(args: Namespace):
    log_dir = parse_path(args.log_dir)
    create_dir(log_dir)
    save_weights = False
    if args.save_dir is not None:
        save_weights = True
        save_dir = parse_path(args.save_dir)
        create_dir(save_dir)
        create_dir(save_dir + 'weights')
        create_dir(save_dir + 'confs')
        create_dir(save_dir + 'normalizers')
    # Datasets
    train_dataloader, val_dataloader, test_dataloader = load_dataset(
        train_node_dir=args.train_node_dir,
        val_node_dir=args.validation_node_dir,
        test_node_dir=args.test_node_dir,
        bsize=args.batch_size,
        remove_prior=args.disable_prior,
        remove_morph=args.disable_morph_feats
    )
    # Configurations
    confs = generate_configurations(args.num_confs, args.model_name)
    create_results_file(args.save_file, args.num_classes)
    num_feats = 18 + (1 if args.num_classes == 2 else args.num_classes)
    if args.disable_morph_feats:
        num_feats -= 18
    if args.disable_prior:
        num_feats -= (1 if args.num_classes == 2 else args.num_classes)
    if num_feats == 0:
        num_feats = 1
    if args.num_workers > 0:
        with ThreadPoolExecutor(max_workers=args.num_workers) as executor:
            futures = []
            for conf in confs:
                future = executor.submit(
                    train_one_conf,
                    args, conf, train_dataloader, val_dataloader, test_dataloader, log_dir, save_weights, save_dir, args.num_classes, num_feats
                )
                futures.append(future)
            for future in futures:
                if args.num_classes == 2:
                    test_f1, test_acc, test_auc, test_perc_err, test_ece, model, conf = future.result()
                    append_results(args.save_file, test_f1, test_acc, test_auc, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'], test_ece, test_perc_err)
                else:
                    test_micro, test_macro, test_weighted, test_ece, model, conf = future.result()
                    append_results(args.save_file, test_micro, test_macro, test_weighted, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'], test_ece)
                if save_weights:
                    save_model(save_dir, model, conf, train_dataloader.dataset.get_normalizers(), 'last_')
    else:
        for conf in confs:
            if args.num_classes == 2:
                test_f1, test_acc, test_auc, test_perc_err, test_ece, model, conf = train_one_conf(
                    args, conf, train_dataloader, val_dataloader, test_dataloader, log_dir, save_weights, save_dir, args.num_classes, num_feats
                )
                append_results(args.save_file, test_f1, test_acc, test_auc, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'], test_ece, test_perc_err)
            else:
                test_micro, test_macro, test_weighted, test_ece, model, conf = train_one_conf(
                    args, conf, train_dataloader, val_dataloader, test_dataloader, log_dir, save_weights, save_dir, args.num_classes, num_feats
                )
                append_results(args.save_file, test_micro, test_macro, test_weighted, conf['NUM_LAYERS'], conf['DROPOUT'], conf['NORM_TYPE'], test_ece)
            if save_weights:
                save_model(save_dir, model, conf, train_dataloader.dataset.get_normalizers(), 'last_')


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main(args)



"""
Computes centroids csv from png <-> csv labels.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from tqdm import tqdm
from ..utils.preprocessing import read_labels, parse_path, create_dir, get_names, save_centroids, get_centroid_by_id
import numpy as np
import pandas as pd
from typing import List, Tuple


def extract_centroids(img: np.ndarray, csv: pd.DataFrame) -> List[Tuple[int, int, int]]:
    """
    Extracts the centroids of cells from a labeled image. The third coordinate is the class.

    :param img: A 2D NumPy array representing the labeled image. Each unique non-zero value represents a different cell.
    :type img: np.ndarray
    :param csv: A pandas DataFrame representing the labels. Just two colums, one for id and another for the class.
    :type csv: pd.DataFrame
    :return: A list of tuples containing the x and y coordinates of the centroid, and the cell class.
    :rtype: List[Tuple[int,int,int]]
    """
    centroids = []
    for _, row in csv.iterrows():
        x, y = get_centroid_by_id(img, row.id)
        if x == -1:
            continue
        centroids.append((x, y, row.label))
    return centroids


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--png-dir', type=str, required=True,
                        help='Path to png files.')
    parser.add_argument('--csv-dir', type=str, required=True,
                        help='Path to csv files.')
    parser.add_argument('--output-path', type=str, required=True,
                        help='Path to save files.')
    return parser


def main_with_args(args):
    png_dir = parse_path(args.png_dir)
    csv_dir = parse_path(args.csv_dir)
    output_path = parse_path(args.output_path)
    create_dir(output_path)

    names = get_names(png_dir, '.GT_cells.png')
    for name in tqdm(names):
        img, csv = read_labels(name, png_dir, csv_dir)
        centroids = extract_centroids(img, csv)
        save_centroids(centroids, output_path, name)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import shutil
import numpy as np

from tumourkit.utils.preprocessing import parse_path, get_names
from tumourkit.segmentation.hovernet.extract_patches import save_npy, to4dim, to5dim, create_dir

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PNGCSV_DIR = parse_path(TEST_DIR) + 'pngcsv/'
ORIG_DIR = parse_path(TEST_DIR) + 'tiles/'
OUT_DIR = parse_path(TEST_DIR) + 'npy_files/'

@pytest.mark.parametrize('shape', ['518', '270'])
def test_hovernet_patches_5dim(shape):
    if os.path.exists(OUT_DIR):
        shutil.rmtree(OUT_DIR)
    create_dir(OUT_DIR)
    img_list = get_names(PNGCSV_DIR, '.GT_cells.png')[:5]
    save_npy(
        img_list, OUT_DIR, to5dim, False,
        ORIG_DIR, PNGCSV_DIR, PNGCSV_DIR,
        shape)
    if shape == '518':
        check_shape = (518,518,5)
    elif shape == '270':
        check_shape = (270,270,5)
    npy_list = get_names(OUT_DIR, '.npy')
    for file in npy_list:
        arr = np.load(OUT_DIR + file + '.npy')
        if arr.shape != check_shape:
            assert False
    assert True
    shutil.rmtree(OUT_DIR)

@pytest.mark.parametrize('shape', ['518', '270'])
def test_hovernet_patches_4dim(shape):
    if os.path.exists(OUT_DIR):
        shutil.rmtree(OUT_DIR)
    create_dir(OUT_DIR)
    img_list = get_names(PNGCSV_DIR, '.GT_cells.png')[:5]
    save_npy(
        img_list, OUT_DIR, to4dim, False,
        ORIG_DIR, PNGCSV_DIR, PNGCSV_DIR,
        shape)
    if shape == '518':
        check_shape = (518,518,4)
    elif shape == '270':
        check_shape = (270,270,4)
    npy_list = get_names(OUT_DIR, '.npy')
    for file in npy_list:
        arr = np.load(OUT_DIR + file + '.npy')
        if arr.shape != check_shape:
            assert False
    assert True
    shutil.rmtree(OUT_DIR)


import cv2
import numpy as np
from scipy.ndimage import measurements
from scipy.ndimage.morphology import binary_fill_holes
from skimage.segmentation import watershed
from ...misc.utils import get_bounding_box, remove_small_objects
import warnings


def noop(*args, **kargs):
    pass


warnings.warn = noop


####
def __proc_np_hv(pred):
    """Process Nuclei Prediction with XY Coordinate Map.

    Args:
        pred: prediction output, assuming 
              channel 0 contain probability map of nuclei
              channel 1 containing the regressed X-map
              channel 2 containing the regressed Y-map

    """
    pred = np.array(pred, dtype=np.float32)

    blb_raw = pred[..., 0]
    h_dir_raw = pred[..., 1]
    v_dir_raw = pred[..., 2]

    # processing
    blb = np.array(blb_raw >= 0.5, dtype=np.int32)

    blb = measurements.label(blb)[0]
    blb = remove_small_objects(blb, min_size=10)
    blb[blb > 0] = 1  # background is 0 already

    h_dir = cv2.normalize(
        h_dir_raw, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
    )
    v_dir = cv2.normalize(
        v_dir_raw, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
    )

    sobelh = cv2.Sobel(h_dir, cv2.CV_64F, 1, 0, ksize=21)
    sobelv = cv2.Sobel(v_dir, cv2.CV_64F, 0, 1, ksize=21)

    sobelh = 1 - (
        cv2.normalize(
            sobelh, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
        )
    )
    sobelv = 1 - (
        cv2.normalize(
            sobelv, None, alpha=0, beta=1, norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F
        )
    )

    overall = np.maximum(sobelh, sobelv)
    overall = overall - (1 - blb)
    overall[overall < 0] = 0

    dist = (1.0 - overall) * blb
    ## nuclei values form mountains so inverse to get basins
    dist = -cv2.GaussianBlur(dist, (3, 3), 0)

    overall = np.array(overall >= 0.4, dtype=np.int32)

    marker = blb - overall
    marker[marker < 0] = 0
    marker = binary_fill_holes(marker).astype("uint8")
    kernel = cv2.getStructuringElement(cv2.MORPH_ELLIPSE, (5, 5))
    marker = cv2.morphologyEx(marker, cv2.MORPH_OPEN, kernel)
    marker = measurements.label(marker)[0]
    marker = remove_small_objects(marker, min_size=10)

    proced_pred = watershed(dist, markers=marker, mask=blb)

    return proced_pred


####
def process(pred_map, nr_types=None, return_centroids=False):
    """Post processing script for image tiles.

    Args:
        pred_map: commbined output of tp, np and hv branches, in the same order
        nr_types: number of types considered at output of nc branch
        overlaid_img: img to overlay the predicted instances upon, `None` means no
        type_colour (dict) : `None` to use random, else overlay instances of a type to colour in the dict
        output_dtype: data type of output
    
    Returns:
        pred_inst:     pixel-wise nuclear instance segmentation prediction
        pred_type_out: pixel-wise nuclear type prediction 

    """
    if nr_types is not None:
        pred_type = pred_map[..., :1]
        pred_inst = pred_map[..., 1:]
        pred_type = pred_type.astype(np.int32)
    else:
        pred_inst = pred_map

    pred_inst = np.squeeze(pred_inst)
    pred_inst = __proc_np_hv(pred_inst)

    inst_info_dict = None
    if return_centroids or nr_types is not None:
        inst_id_list = np.unique(pred_inst)[1:]  # exlcude background
        inst_info_dict = {}
        for inst_id in inst_id_list:
            inst_map = pred_inst == inst_id
            # TODO: chane format of bbox output
            rmin, rmax, cmin, cmax = get_bounding_box(inst_map)
            inst_bbox = np.array([[rmin, cmin], [rmax, cmax]])
            inst_map = inst_map[
                inst_bbox[0][0] : inst_bbox[1][0], inst_bbox[0][1] : inst_bbox[1][1]
            ]
            inst_map = inst_map.astype(np.uint8)
            inst_moment = cv2.moments(inst_map)
            inst_contour = cv2.findContours(
                inst_map, cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
            )
            # * opencv protocol format may break
            inst_contour = np.squeeze(inst_contour[0][0].astype("int32"))
            # < 3 points dont make a contour, so skip, likely artifact too
            # as the contours obtained via approximation => too small or sthg
            if inst_contour.shape[0] < 3:
                continue
            if len(inst_contour.shape) != 2:
                continue # ! check for trickery shape
            inst_centroid = [
                (inst_moment["m10"] / inst_moment["m00"]),
                (inst_moment["m01"] / inst_moment["m00"]),
            ]
            inst_centroid = np.array(inst_centroid)
            inst_contour[:, 0] += inst_bbox[0][1]  # X
            inst_contour[:, 1] += inst_bbox[0][0]  # Y
            inst_centroid[0] += inst_bbox[0][1]  # X
            inst_centroid[1] += inst_bbox[0][0]  # Y
            inst_info_dict[inst_id] = {  # inst_id should start at 1
                "bbox": inst_bbox,
                "centroid": inst_centroid,
                "contour": inst_contour,
                "type_prob": None,
                "prob1": None, # Added probability of positive
                "type": None,
            }
            if nr_types is not None and int(nr_types) > 3:
                for k in range(1, int(nr_types)):
                    inst_info_dict[inst_id]["prob" + str(k)] = None # Initialized probabilities of all classes

    if nr_types is not None:
        #### * Get class of each instance id, stored at index id-1
        for inst_id in list(inst_info_dict.keys()):
            rmin, cmin, rmax, cmax = (inst_info_dict[inst_id]["bbox"]).flatten()
            inst_map_crop = pred_inst[rmin:rmax, cmin:cmax]
            inst_type_crop = pred_type[rmin:rmax, cmin:cmax]
            inst_map_crop = (
                inst_map_crop == inst_id
            )  # TODO: duplicated operation, may be expensive
            inst_type = inst_type_crop[inst_map_crop]
            type_list, type_pixels = np.unique(inst_type, return_counts=True)
            type_list = list(zip(type_list, type_pixels))
            type_list = sorted(type_list, key=lambda x: x[1], reverse=True)
            inst_type = type_list[0][0]
            if inst_type == 0:  # ! pick the 2nd most dominant if exist
                if len(type_list) > 1:
                    inst_type = type_list[1][0]
            type_dict = {v[0]: v[1] for v in type_list}
            type_prob = type_dict[inst_type] / (np.sum(inst_map_crop) + 1.0e-6)
            inst_info_dict[inst_id]["type"] = int(inst_type)
            inst_info_dict[inst_id]["type_prob"] = float(type_prob)
            # Compute probability of positive (binary)
            if not 2 in type_dict and not 1 in type_dict:
                prob1 = 0
            elif not 1 in type_dict:
                prob1 = 1
            elif not 2 in type_dict:
                prob1 = 0
            else:
                prob1 = type_dict[2] / (type_dict[1] + type_dict[2])
            inst_info_dict[inst_id]["prob1"] = float(prob1)
            # Multi-class probabilities
            if int(nr_types) > 3:
                total_sum = np.sum(list(type_dict.values()))
                if 0 in type_dict:
                    total_sum -= type_dict[0]
                for k in range(1, int(nr_types)):
                    if not k in type_dict:
                        type_count = 0
                    else:
                        type_count = type_dict[k]
                    inst_info_dict[inst_id]["prob" + str(k)] = float(type_count / total_sum)

    # print('here')
    # ! WARNING: ID MAY NOT BE CONTIGUOUS
    # inst_id in the dict maps to the same value in the `pred_inst`
    return pred_inst, inst_info_dict



"""
Main HoVer-Net training script.
"""
import cv2
cv2.setNumThreads(0)
import glob
import inspect
import json
import os
import numpy as np
import matplotlib.pyplot as plt
import torch
from tensorboardX import SummaryWriter
from torch.nn import DataParallel
from torch.utils.data import DataLoader
from .config import Config
from .dataloader.train_loader import FileLoader
from .misc.utils import rm_n_mkdir
from .run_utils.engine import RunEngine
from .run_utils.utils import (
    check_manual_seed,
    colored,
    convert_pytorch_checkpoint
)
import argparse


def worker_init_fn(worker_id):
    worker_info = torch.utils.data.get_worker_info()
    worker_seed = torch.randint(0, 2 ** 32, (1,))[0].cpu().item() + worker_id
    worker_info.dataset.setup_augmentor(worker_id, worker_seed)
    return


class TrainManager(Config):
    """Either used to view the dataset or to initialise the main training loop."""


    def view_dataset(self, save_name, mode="train"):
        self.nr_gpus = 1
        check_manual_seed(self.seed)
        phase_list = self.model_config["phase_list"][0]
        target_info = phase_list["target_info"]
        prep_func, prep_kwargs = target_info["viz"]
        dataloader = self._get_datagen(2, mode, target_info["gen"])
        for i, batch_data in enumerate(dataloader):  
            batch_data = {k: v.numpy() for k, v in batch_data.items()}
            viz = prep_func(batch_data, is_batch=True, **prep_kwargs)
            plt.imshow(viz)
            plt.savefig(save_name + str(i) + '.png')
        self.nr_gpus = -1
        return


    def _get_datagen(self, batch_size, run_mode, target_gen, nr_procs=0, fold_idx=0):
        nr_procs = nr_procs if not self.debug else 0

        # ! Hard assumption on file type
        file_list = []
        if run_mode == "train":
            data_dir_list = self.train_dir_list
        else:
            data_dir_list = self.valid_dir_list
        for dir_path in data_dir_list:
            file_list.extend(glob.glob("%s/*.npy" % dir_path))
        file_list.sort()  # to always ensure same input ordering

        assert len(file_list) > 0, (
            "No .npy found for `%s`, please check `%s` in `config.py`"
            % (run_mode, "%s_dir_list" % run_mode)
        )
        print("Dataset %s: %d" % (run_mode, len(file_list)))
        input_dataset = FileLoader(
            file_list,
            mode=run_mode,
            with_type=self.type_classification,
            setup_augmentor=nr_procs == 0,
            target_gen=target_gen,
            **self.shape_info[run_mode]
        )

        dataloader = DataLoader(
            input_dataset,
            num_workers=nr_procs,
            batch_size=batch_size * max(1, self.nr_gpus),
            shuffle=run_mode == "train",
            drop_last=run_mode == "train",
            worker_init_fn=worker_init_fn,
        )
        return dataloader


    def run_once(self, opt, run_engine_opt, log_dir, prev_log_dir=None, fold_idx=0):
        """Simply run the defined run_step of the related method once."""
        check_manual_seed(self.seed)

        log_info = {}
        if self.logging:
            rm_n_mkdir(log_dir)

            tfwriter = SummaryWriter(log_dir=log_dir)
            json_log_file = log_dir + "/stats.json"
            with open(json_log_file, "w") as json_file:
                json.dump({}, json_file)  # create empty file
            log_info = {
                "json_file": json_log_file,
                "tfwriter": tfwriter,
            }

        loader_dict = {}
        for runner_name, runner_opt in run_engine_opt.items():
            loader_dict[runner_name] = self._get_datagen(
                opt["batch_size"][runner_name],
                runner_name,
                opt["target_info"]["gen"],
                nr_procs=runner_opt["nr_procs"],
                fold_idx=fold_idx,
            )
        
        def get_last_chkpt_path(prev_phase_dir, net_name):
            stat_file_path = prev_phase_dir + "/stats.json"
            with open(stat_file_path) as stat_file:
                info = json.load(stat_file)
            epoch_list = [int(v) for v in info.keys()]
            last_chkpts_path = "%s/%s_epoch=%d.tar" % (
                prev_phase_dir,
                net_name,
                max(epoch_list),
            )
            return last_chkpts_path

        net_run_info = {}
        net_info_opt = opt["run_info"]
        for net_name, net_info in net_info_opt.items():
            assert inspect.isclass(net_info["desc"]) or inspect.isfunction(
                net_info["desc"]
            ), "`desc` must be a Class or Function which instantiate NEW objects !!!"
            net_desc = net_info["desc"]()
            pretrained_path = net_info["pretrained"]
            if pretrained_path is not None:
                if pretrained_path == -1:
                    # * depend on logging format so may be broken if logging format has been changed
                    pretrained_path = get_last_chkpt_path(prev_log_dir, net_name)
                    net_state_dict = torch.load(pretrained_path)["desc"]
                else:
                    chkpt_ext = os.path.basename(pretrained_path).split(".")[-1]
                    if chkpt_ext == "npz":
                        net_state_dict = dict(np.load(pretrained_path))
                        net_state_dict = {
                            k: torch.from_numpy(v) for k, v in net_state_dict.items()
                        }
                    elif chkpt_ext == "tar":  # ! assume same saving format we desire
                        net_state_dict = torch.load(pretrained_path)["desc"]
                        # TODO: make this optional
                        del net_state_dict['decoder.tp.u0.conv.weight']
                        del net_state_dict['decoder.tp.u0.conv.bias']

                colored_word = colored(net_name, color="red", attrs=["bold"])
                print(
                    "Model `%s` pretrained path: %s" % (colored_word, pretrained_path)
                )

                net_state_dict = convert_pytorch_checkpoint(net_state_dict)
                load_feedback = net_desc.load_state_dict(net_state_dict, strict=False)
                # * uncomment for your convenience
                print("Missing Variables: \n", load_feedback[0])
                print("Detected Unknown Variables: \n", load_feedback[1])

            net_desc = DataParallel(net_desc)
            if self.nr_gpus > 0:
                net_desc = net_desc.to("cuda")
            optimizer, optimizer_args = net_info["optimizer"]
            optimizer = optimizer(net_desc.parameters(), **optimizer_args)
            scheduler = net_info["lr_scheduler"](optimizer)
            net_run_info[net_name] = {
                "use_cpu": net_info["use_cpu"],
                "desc": net_desc,
                "optimizer": optimizer,
                "lr_scheduler": scheduler,
                # TODO: standardize API for external hooks
                "extra_info": net_info["extra_info"],
            }

        # parsing the running engine configuration
        assert (
            "train" in run_engine_opt
        ), "No engine for training detected in description file"

        # initialize runner and attach callback afterward
        # * all engine shared the same network info declaration
        runner_dict = {}
        for runner_name, runner_opt in run_engine_opt.items():
            runner_dict[runner_name] = RunEngine(
                dataloader=loader_dict[runner_name],
                engine_name=runner_name,
                run_step=runner_opt["run_step"],
                run_info=net_run_info,
                log_info=log_info,
            )

        for runner_name, runner in runner_dict.items():
            callback_info = run_engine_opt[runner_name]["callbacks"]
            for event, callback_list, in callback_info.items():
                for callback in callback_list:
                    if callback.engine_trigger:
                        triggered_runner_name = callback.triggered_engine_name
                        callback.triggered_engine = runner_dict[triggered_runner_name]
                    runner.add_event_handler(event, callback)

        # retrieve main runner
        main_runner = runner_dict["train"]
        main_runner.state.logging = self.logging
        main_runner.state.log_dir = log_dir
        # start the run loop
        main_runner.run(opt["nr_epochs"])

        print("\n")
        print("########################################################")
        print("########################################################")
        print("\n")
        return


    def run(self):
        self.nr_gpus = torch.cuda.device_count()
        print('Detect #GPUS: %d' % self.nr_gpus)

        phase_list = self.model_config["phase_list"]
        engine_opt = self.model_config["run_engine"]

        prev_save_path = None
        for phase_idx, phase_info in enumerate(phase_list):
            if len(phase_list) == 1:
                save_path = self.log_dir
            else:
                save_path = self.log_dir + "/%02d/" % (phase_idx)
            self.run_once(
                phase_info, engine_opt, save_path, prev_log_dir=prev_save_path
            )
            prev_save_path = save_path


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--gpu', type=str, default='')
    parser.add_argument('--view', type=str)
    parser.add_argument('--save-name', type=str)
    parser.add_argument('--log-dir', type=str, required=True)
    parser.add_argument('--train-dir', type=str, required=True)
    parser.add_argument('--valid-dir', type=str, required=True)
    parser.add_argument('--shape', type=str, required=True, choices=['270', '518'])
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main_with_args(args):
    trainer = TrainManager(
        args.shape, args.log_dir, args.train_dir, args.valid_dir,
        args.pretrained_path, use_cpu=(args.gpu == ''), num_classes=args.num_classes
    )

    if args.view is not None:
        assert args.view == "train" or args.view == "valid", 'Use "train" or "valid" for --view.'
        assert args.save_name is not None, 'In view mode, an image name must be provided.'
        trainer.view_dataset(save_name=args.save_name, mode=args.view)
    else:
        os.environ["CUDA_VISIBLE_DEVICES"] = args.gpu
        trainer.run()

def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)
    



"""
Given .nodes.csv prediction files and .centroids.csv GT files,
replaces target value of .nodes.csv with GT ones.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pandas as pd
import numpy as np
from tqdm import tqdm
import argparse
from argparse import Namespace
import os
from ..utils.preprocessing import parse_path, create_dir, get_names, read_centroids, save_graph
from ..utils.nearest import generate_tree, find_nearest


def merge_labels(graph: pd.DataFrame, centroids: np.ndarray) -> pd.DataFrame:
    """
    Matches nodes in the graph to centroids and updates their labels.

    Given a graph DataFrame and a centroids array, this function finds 1-1 matchings
    between nodes in the graph and centroids based on their coordinates.
    It substitutes the labels of the graph nodes with the labels from the centroids.
    The matching is performed using KD-trees for efficient nearest neighbor search.

    :param graph: The graph DataFrame containing nodes with labels.
    :type graph: pd.DataFrame
    :param centroids: The centroids array containing coordinates and labels.
    :type centroids: np.ndarray
    :return: The updated graph DataFrame with matched labels.
    :rtype: pd.DataFrame
    """
    assert len(centroids) > 0, 'GT must contain at least one cell.'
    graph = graph.copy()
    graph['prob1'] = graph['class'] - 1
    gt_tree = generate_tree(centroids[:, :2])
    pred_centroids = graph[['X', 'Y']].to_numpy(dtype=int)
    pred_tree = generate_tree(pred_centroids)
    for point_id, point in enumerate(centroids):
        if point[2] == -1:
            continue
        closest_id = find_nearest(point[:2], pred_tree)
        closest = graph.loc[closest_id, ['X', 'Y', 'class']]
        if closest[2] == -1:
            continue
        if point_id == find_nearest(closest[:2], gt_tree):
            graph.loc[closest_id, 'class'] = point[2]  # 1-1 matchings
    for point_id, point in enumerate(pred_centroids):
        closest_id = find_nearest(point[:2], gt_tree)
        closest = centroids[closest_id]
        if point_id != find_nearest(closest[:2], pred_tree):
            graph.drop(point_id, inplace=True)  # Remove prediction not matched in GT
    return graph


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '--graph-dir', type=str, required=True,
        help='Path to folder containing prediction .nodes.csv files.'
    )
    parser.add_argument(
        '--centroids-dir', type=str, required=True,
        help='Path to folder containing .centroids.csv files.'
    )
    parser.add_argument(
        '--output-dir', type=str, required=True,
        help='Path to folder where to save new .nodes.csv. If same as --graph-dir, overwrites files.'
    )
    return parser


def main_with_args(args: Namespace) -> None:
    graph_dir = parse_path(args.graph_dir)
    centroids_dir = parse_path(args.centroids_dir)
    output_dir = parse_path(args.output_dir)
    create_dir(output_dir)
    names = sorted(get_names(centroids_dir, '.centroids.csv'))
    for name in tqdm(names):
        try:
            centroids = read_centroids(name, centroids_dir)
            graph = pd.read_csv(os.path.join(graph_dir, name + '.nodes.csv'))
            graph = merge_labels(graph, centroids)
            save_graph(graph, os.path.join(output_dir, name + '.nodes.csv'))
        except FileNotFoundError:
            continue
    return


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main_with_args(args)



"""
Script to generate predictions from given GNN model.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
from typing import Dict, Tuple, Any
from ..utils.preprocessing import parse_path, create_dir
import torch
from torch import nn
import torch.nn.functional as F
from .train_graphs import load_model
from .read_graph import GraphDataset
import dgl
from dgl.dataloading import GraphDataLoader
import json
import pickle
import numpy as np
import pandas as pd
import argparse
import os


def load_saved_model(weights_path: str, conf_path: str, num_classes: int, num_feats: int) -> nn.Module:
    """
    Loads a saved model from the given weights_path and conf_path.

    :param weights_path: The path to the saved weights file.
    :type weights_path: str
    :param conf_path: The path to the configuration file.
    :type conf_path: str
    :param num_classes: The number of classes for the model.
    :type num_classes: int
    :param num_feats: The number of features for the model.
    :type num_feats: int
    :return: The loaded model.
    :rtype: nn.Module
    """
    state_dict = torch.load(weights_path, map_location='cpu')
    with open(conf_path, 'r') as f:
        conf = json.load(f)
    model = load_model(conf, num_classes, num_feats)
    model.load_state_dict(state_dict)
    return model


def load_normalizer(norm_path: str) -> Tuple[Any]:
    """
    Loads the normalizers used in training from the given norm_path.

    :param norm_path: The path to the saved normalizers file.
    :type norm_path: str
    :return: The loaded normalizers.
    :rtype: Tuple[Any]
    """
    with open(norm_path, 'rb') as f:
        normalizers = pickle.load(f)
    return normalizers


def run_inference(model: nn.Module, loader: GraphDataLoader, device: str, num_classes: int) -> Dict[str, np.ndarray]:
    """
    Runs inference using the specified model on the provided data loader.

    :param model: The model used for inference.
    :type model: nn.Module
    :param loader: The graph data loader.
    :type loader: GraphDataLoader
    :param device: The device used for inference (e.g., 'cpu' or 'cuda').
    :type device: str
    :param num_classes: The number of classes.
    :type num_classes: int
    :return: The probabilities for all the nodes.
    :rtype: Dict[str, np.ndarray]
    """
    probs = {}
    for g, name in loader:
        # self-loops
        g = dgl.remove_self_loop(g)
        g = dgl.add_self_loop(g)
        # data
        features = g.ndata['X'].to(device)
        # Forward
        logits = model(g, features)
        if num_classes == 2:
            prob = F.softmax(logits, dim=1).detach().numpy()[:, 1].reshape(-1, 1)
        else:
            prob = F.softmax(logits, dim=1).detach().numpy()
        probs[name[0]] = prob
    return probs


def save_probs(probs: Dict[str, np.ndarray], node_dir: str, output_dir: str, num_classes: int) -> None:
    """
    Saves the probabilities in .nodes.csv files by appending a column to the original .nodes.csv file.

    :param probs: The probabilities for each graph.
    :type probs: Dict[str, np.ndarray]
    :param node_dir: The directory containing the original .nodes.csv files.
    :type node_dir: str
    :param output_dir: The directory where the updated .nodes.csv files will be saved.
    :type output_dir: str
    :param num_classes: The number of classes.
    :type num_classes: int
    """
    for name, prob in probs.items():
        orig = pd.read_csv(os.path.join(node_dir, name))
        if num_classes == 2:
            orig['prob1'] = prob
        else:
            for k in range(1, num_classes + 1):
                orig['prob' + str(k)] = prob[:, (k - 1)]
        orig.to_csv(output_dir + name, index=False)


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--node-dir', type=str, required=True,
                        help='Folder containing .nodes.csv')
    parser.add_argument('--output-dir', type=str, required=True,
                        help='Folder to save .nodes.csv containing probabilities.')
    parser.add_argument('--weights', type=str, required=True,
                        help='Path to model weights.')
    parser.add_argument('--conf', type=str, required=True,
                        help='Configuration file for the model.')
    parser.add_argument('--normalizers', type=str, required=True,
                        help='Path to normalizer objects for the model.')
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    parser.add_argument('--disable-prior', action='store_true', help='If True, remove hovernet probabilities from node features.')
    parser.add_argument('--disable-morph-feats', action='store_true', help='If True, remove morphological features from node features.')
    return parser


def main_with_args(args):
    node_dir = parse_path(args.node_dir)
    output_dir = parse_path(args.output_dir)
    create_dir(output_dir)
    normalizers = load_normalizer(args.normalizers)
    eval_dataset = GraphDataset(
        node_dir=node_dir, return_names=True, is_inference=True,
        max_dist=200, max_degree=10, normalizers=normalizers,
        remove_morph=args.disable_morph_feats, remove_prior=args.disable_prior)
    eval_dataloader = GraphDataLoader(eval_dataset, batch_size=1, shuffle=False)
    num_feats = 18 + (1 if args.num_classes == 2 else args.num_classes)
    if args.disable_morph_feats:
        num_feats -= 18
    if args.disable_prior:
        num_feats -= (1 if args.num_classes == 2 else args.num_classes)
    if num_feats == 0:
        num_feats = 1
    model = load_saved_model(args.weights, args.conf, args.num_classes, num_feats)
    model.eval()
    probs = run_inference(model, eval_dataloader, 'cpu', args.num_classes)
    save_probs(probs, node_dir, output_dir, args.num_classes)


def main():
    parser = _create_parser()
    args = parser.parse_args()
    main(args)



import json
from importlib import import_module
import matplotlib.pyplot as plt
import numpy as np
import torch
from ..run_utils.utils import convert_pytorch_checkpoint


####
class InferManager(object):
    def __init__(self, **kwargs):
        self.run_step = None
        for variable, value in kwargs.items():
            self.__setattr__(variable, value)
        self.__load_model()
        self.nr_types = self.method["model_args"]["nr_types"]
        # create type info name and colour

        # default
        self.type_info_dict = {
            None: ["no label", [0, 0, 0]],
        }

        if self.nr_types is not None and self.type_info_path is not None:
            self.type_info_dict = json.load(open(self.type_info_path, "r"))
            self.type_info_dict = {
                int(k): (v[0], tuple(v[1])) for k, v in self.type_info_dict.items()
            }
            # availability check
            for k in range(self.nr_types):
                if k not in self.type_info_dict:
                    assert False, "Not detect type_id=%d defined in json." % k

        if self.nr_types is not None and self.type_info_path is None:
            cmap = plt.get_cmap("hot")
            colour_list = np.arange(self.nr_types, dtype=np.int32)
            colour_list = (cmap(colour_list)[..., :3] * 255).astype(np.uint8)
            # should be compatible out of the box wrt qupath
            self.type_info_dict = {
                k: (str(k), tuple(v)) for k, v in enumerate(colour_list)
            }
        return

    def __load_model(self):
        """Create the model, load the checkpoint and define
        associated run steps to process each data batch.
        
        """
        model_desc = import_module(".segmentation.hovernet.models.hovernet.net_desc", package="tumourkit")
        model_creator = getattr(model_desc, "create_model")

        net = model_creator(**self.method["model_args"])
        if self.nr_gpus > 0:
            saved_state_dict = torch.load(self.method["model_path"])["desc"]
        else:
            saved_state_dict = torch.load(self.method["model_path"], map_location=torch.device('cpu'))["desc"]
        saved_state_dict = convert_pytorch_checkpoint(saved_state_dict)

        net.load_state_dict(saved_state_dict, strict=True)
        net = torch.nn.DataParallel(net)
        if self.nr_gpus > 0:
            net = net.to("cuda")

        module_lib = import_module(".segmentation.hovernet.models.hovernet.run_desc", package="tumourkit")
        run_step = getattr(module_lib, "infer_step")
        self.run_step = lambda input_batch: run_step(input_batch, net, use_cpu=self.nr_gpus==0)

        module_lib = import_module(".segmentation.hovernet.models.hovernet.post_proc", package="tumourkit")
        self.post_proc_func = getattr(module_lib, "process")
        return

    def __save_json(self, path, old_dict, mag=None):
        new_dict = {}
        for inst_id, inst_info in old_dict.items():
            new_inst_info = {}
            for info_name, info_value in inst_info.items():
                # convert to jsonable
                if isinstance(info_value, np.ndarray):
                    info_value = info_value.tolist()
                new_inst_info[info_name] = info_value
            new_dict[int(inst_id)] = new_inst_info

        json_dict = {"mag": mag, "nuc": new_dict}  # to sync the format protocol
        with open(path, "w") as handle:
            json.dump(json_dict, handle)
        return new_dict



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest
import os
import pandas as pd
import numpy as np

from tumourkit.preprocessing.png2graph import png2graph
from tumourkit.utils.preprocessing import get_names, parse_path, read_png, read_graph
from tumourkit.utils.nearest import generate_tree, find_nearest_dist_idx

TEST_DIR = os.path.dirname(os.path.abspath(__file__))
PNGCSV_DIR = parse_path(TEST_DIR) + 'pngcsv/'
GRAPHS_DIR = parse_path(TEST_DIR) + 'graphs/'
THRESHOLD = 5


def check_equal(a: pd.DataFrame, b: pd.DataFrame) -> bool:
    a_tree = generate_tree(a[['X', 'Y']].to_numpy())
    mean_dist = 0
    for point in b[['X', 'Y']].to_numpy():
        closest_dist, _ = find_nearest_dist_idx(point, a_tree)
        mean_dist += closest_dist
    mean_dist /= len(b)
    return mean_dist < THRESHOLD


@pytest.mark.parametrize("name", get_names(PNGCSV_DIR, '.GT_cells.png'))
def test_png2graph(name):
    png_file = read_png(name, PNGCSV_DIR)
    false_orig = np.ones((png_file.shape[0], png_file.shape[1], 3), dtype=np.uint8)
    graph = png2graph(false_orig, png_file)
    ref_graph = read_graph(name, GRAPHS_DIR)
    if not check_equal(graph, ref_graph):
        assert False, name
    assert(True)



"""
Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import pytest

from tumourkit.utils.postprocessing import remove_idx



@pytest.mark.parametrize("a,a_idx,out", [
    ([0,0,[[0,0],[1,1],[2,2],[3,3]]], [2], (0,0,[[3,3],[0,0],[1,1]])),
    (
        [0,0,[[0,0],[1,1],[2,2],[3,3],[4,4],[5,5],[6,6],[7,7],[8,8],[9,9],[10,10]]], 
        [9,10,0,1,2,6,8], 
        (0,0,[[3,3],[4,4],[5,5],[6,6],[7,7]]),
    ),
    (
        [0,0,[[0,0],[1,1],[2,2],[3,3],[4,4],[5,5],[6,6],[7,7],[8,8],[9,9],[10,10],[11,11]]], 
        [3,2,0,4,11,5,6,9],
        (0,0,[[7,7],[8,8],[9,9],[10,10],[11,11],[0,0],[1,1]]), 
    )
])
def test_conversion(a,a_idx,out):
    pred = remove_idx(a, a_idx)
    assert(pred==out)



"""
Training pipeline.

Copyright (C) 2023  Jose Pérez Cano

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.

Contact information: joseperez2000@hotmail.es
"""
import argparse
from argparse import Namespace
from .preprocessing import geojson2pngcsv_main, pngcsv2graph_main, hovernet2geojson_main, pngcsv2centroids_main
from .segmentation import pngcsv2npy
import os
import logging
from logging import Logger
from .segmentation import hov_train, hov_infer
from .utils.preprocessing import get_names
import shutil
from .postprocessing import join_graph_gt_main, join_hovprob_graph_main
from .classification import train_gnn


def run_preproc_pipe(args: Namespace, logger: Logger) -> None:
    """
    Runs the preprocessing pipeline to convert the gson format to other formats.

    This function performs the following steps:
        1. Converts the geojson files to pngcsv format.
        2. Converts the pngcsv files to npy format.

    :param args: The arguments for the preprocessing pipeline.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    for split in ['train', 'validation', 'test']:
        logger.info(f'Parsing {split} split')
        newargs = Namespace(
            gson_dir=os.path.join(args.root_dir, 'data', split, 'gson'),
            png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
            csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
            num_classes=args.num_classes,
        )
        geojson2pngcsv_main(newargs)
        newargs = Namespace(
            orig_dir=os.path.join(args.root_dir, 'data', 'orig'),
            png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
            csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
            out_dir=os.path.join(args.root_dir, 'data', split, 'npy'),
            save_example=False, use_labels=True, split=False,
            shape='518'
        )
        pngcsv2npy(newargs)
    return


def run_hov_pipe(args: Namespace, logger: Logger) -> None:
    """
    Trains Hovernet and predicts cell contours in json format.

    This function performs the following steps:
        1. Trains Hovernet on the training data.
        2. Performs inference using the trained Hovernet model on the input images.
        3. Saves the predicted cell contours in json format.

    :param args: The arguments for the Hovernet pipeline.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Starting training.')
    newargs = Namespace(
        gpu=args.gpu, view=None, save_name=None,
        log_dir=os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet'),
        train_dir=os.path.join(args.root_dir, 'data', 'train', 'npy'),
        valid_dir=os.path.join(args.root_dir, 'data', 'validation', 'npy'),
        pretrained_path=args.pretrained_path,
        shape='518',
        num_classes=args.num_classes,
    )
    hov_train(newargs)
    logger.info('Starting inference.')
    newargs = {
        'nr_types': str(args.num_classes + 1),
        'type_info_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', 'type_info.json'),
        'gpu': args.gpu,
        'nr_inference_workers': '0',
        'model_path': os.path.join(args.root_dir, 'weights', 'segmentation', 'hovernet', '01', 'net_epoch=50.tar'),
        'batch_size': '10',
        'shape': '518',
        'nr_post_proc_workers': '0',
        'model_mode': 'original',
        'help': False
    }
    newsubargs = {
        'input_dir': os.path.join(args.root_dir, 'data', 'orig'),
        'output_dir': os.path.join(args.root_dir, 'data', 'tmp_hov'),
        'draw_dot': False,
        'save_qupath': False,
        'save_raw_map': False,
        'mem_usage': '0.2'
    }
    hov_infer(newargs, newsubargs, 'tile')
    return


def run_postproc_pipe(args: Namespace, logger: Logger) -> None:
    """
    Converts the json format to the graph format containing ground truth (GT) and predictions.

    This function performs the following steps:
        1. Moves json files to their corresponding folders based on the split.
        2. Converts the json format to geojson format.
        3. Converts the geojson format to pngcsv format.
        4. Converts the pngcsv format to nodes.csv format.
        5. Extracts centroids from ground truth (GT) data.
        6. Adds GT labels to the nodes.csv file.
        7. Adds Hovernet predictions to the nodes.csv file.

    :param args: The arguments for the post-processing pipeline.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    logger.info('Moving json files to corresponding folders.')
    tr_files = set(get_names(os.path.join(args.root_dir, 'data', 'train', 'gson'), '.geojson'))
    val_files = set(get_names(os.path.join(args.root_dir, 'data', 'validation', 'gson'), '.geojson'))
    ts_files = set(get_names(os.path.join(args.root_dir, 'data', 'test', 'gson'), '.geojson'))
    json_files = set(get_names(os.path.join(args.root_dir, 'data', 'tmp_hov', 'json'), '.json'))
    for folder_name, split_files in zip(['train', 'validation', 'test'], [tr_files, val_files, ts_files]):
        for file in json_files.intersection(split_files):
            shutil.copy(
                os.path.join(args.root_dir, 'data', 'tmp_hov', 'json', file + '.json'),
                os.path.join(args.root_dir, 'data', folder_name, 'json')
            )
    for split in ['train', 'validation', 'test']:
        logger.info(f'Parsing {split} split')
        logger.info('   From json to geojson.')
        newargs = Namespace(
            json_dir=os.path.join(args.root_dir, 'data', split, 'json'),
            gson_dir=os.path.join(args.root_dir, 'data', split, 'gson_hov'),
            num_classes=args.num_classes
        )
        hovernet2geojson_main(newargs)
        logger.info('   From geojson to pngcsv.')
        newargs = Namespace(
            gson_dir=os.path.join(args.root_dir, 'data', split, 'gson_hov'),
            png_dir=os.path.join(args.root_dir, 'data', split, 'png_hov'),
            csv_dir=os.path.join(args.root_dir, 'data', split, 'csv_hov'),
            num_classes=args.num_classes,
        )
        geojson2pngcsv_main(newargs)
        logger.info('   From pngcsv to nodes.csv.')
        newargs = Namespace(
            png_dir=os.path.join(args.root_dir, 'data', split, 'png_hov'),
            csv_dir=os.path.join(args.root_dir, 'data', split, 'csv_hov'),
            orig_dir=os.path.join(args.root_dir, 'data', 'orig'),
            output_path=os.path.join(args.root_dir, 'data', split, 'graphs', 'raw'),
            num_workers=args.num_workers
        )
        pngcsv2graph_main(newargs)
        logger.info('   Extracting centroids from GT.')
        newargs = Namespace(
            png_dir=os.path.join(args.root_dir, 'data', split, 'png'),
            csv_dir=os.path.join(args.root_dir, 'data', split, 'csv'),
            output_path=os.path.join(args.root_dir, 'data', split, 'centroids')
        )
        pngcsv2centroids_main(newargs)
        logger.info('   Adding GT labels to .nodes.csv.')
        newargs = Namespace(
            graph_dir=os.path.join(args.root_dir, 'data', split, 'graphs', 'raw'),
            centroids_dir=os.path.join(args.root_dir, 'data', split, 'centroids'),
            output_dir=os.path.join(args.root_dir, 'data', split, 'graphs', 'GT')
        )
        join_graph_gt_main(newargs)
        logger.info('   Adding hovernet predictions to .nodes.csv.')
        newargs = Namespace(
            json_dir=os.path.join(args.root_dir, 'data', split, 'json'),
            graph_dir=os.path.join(args.root_dir, 'data', split, 'graphs', 'GT'),
            output_dir=os.path.join(args.root_dir, 'data', split, 'graphs', 'preds'),
            num_classes=args.num_classes,
        )
        join_hovprob_graph_main(newargs, logger)
    return


def run_graph_pipe(args: Namespace, logger: Logger) -> None:
    """
    Trains the graph models.

    This function trains the graph models using the following steps:
        1. Trains the GCN (Graph Convolutional Network) model.
        2. Trains the GAT (Graph Attention Network) model.

    :param args: The arguments for the graph pipeline.
    :type args: Namespace

    :param logger: The logger object used for logging messages.
    :type logger: Logger
    """
    newargs = Namespace(
        train_node_dir=os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir=os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir=os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir=os.path.join(args.root_dir, 'gnn_logs'),
        early_stopping_rounds=10,
        batch_size=20,
        model_name='GCN',
        save_file=os.path.join(args.root_dir, 'gnn_logs', 'gcn_results'),
        num_confs=32,
        save_dir=os.path.join(args.root_dir, 'weights', 'classification', 'gnn'),
        device='cpu' if args.gpu == '' else 'cuda',
        num_workers=args.num_workers,
        checkpoint_iters=-1,
        num_classes=args.num_classes,
        disable_prior=False,
        disable_morph_feats=False,
    )
    train_gnn(newargs)
    newargs = Namespace(
        train_node_dir=os.path.join(args.root_dir, 'data', 'train', 'graphs', 'preds'),
        validation_node_dir=os.path.join(args.root_dir, 'data', 'validation', 'graphs', 'preds'),
        test_node_dir=os.path.join(args.root_dir, 'data', 'test', 'graphs', 'preds'),
        log_dir=os.path.join(args.root_dir, 'gnn_logs'),
        early_stopping_rounds=10,
        batch_size=20,
        model_name='ATT',
        save_file=os.path.join(args.root_dir, 'gnn_logs', 'gat_results'),
        num_confs=32,
        save_dir=os.path.join(args.root_dir, 'weights', 'classification', 'gnn'),
        device='cpu' if args.gpu == '' else 'cuda',
        num_workers=args.num_workers,
        checkpoint_iters=-1,
        num_classes=args.num_classes,
        disable_prior=False,
        disable_morph_feats=False,
    )
    train_gnn(newargs)
    return


def _create_parser():
    parser = argparse.ArgumentParser()
    parser.add_argument('--root-dir', type=str, default='./.internals/', help='Root folder to save data and models.')
    parser.add_argument('--pretrained-path', type=str, help='Path to initial Hovernet weights.')
    parser.add_argument('--gpu', type=str, default='0')
    parser.add_argument('--num-workers', type=int, default=0)
    parser.add_argument('--num-classes', type=int, default=2, help='Number of classes to consider for classification (background not included).')
    return parser


def main():
    parser = _create_parser()
    args = parser.parse_args()

    logger = logging.getLogger('train_pipe')
    logger.setLevel(logging.DEBUG)
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)
    formatter = logging.Formatter('%(name)s - %(levelname)s - %(message)s')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    logger.info('Starting preprocessing pipeline.')
    run_preproc_pipe(args, logger)
    logger.info('Finished preprocessing pipeline.')
    logger.info('Starting Hovernet pipeline.')
    run_hov_pipe(args, logger)
    logger.info('Finished Hovernet pipeline.')
    logger.info('Starting postprocessing pipeline.')
    run_postproc_pipe(args, logger)
    logger.info('Finished postprocessing pipeline.')
    logger.info('Starting graph pipeline.')
    run_graph_pipe(args, logger)
    logger.info('Finished graph pipeline.')
